import numpy as np
from sklearn.inspection import partial_dependence


def local_partial_dependence(local_estimator, X, weight):
    """

    Args:
        local_estimator ():
        X ():
        weight ():

    Returns:

    """
    # TODO: More detailed reason to justify the weighted partial dependence.
    # Care more on the local range.
    # Unweighted points will dominate the tendency which may not be the interested one.
    # Only calculate the local ICE?
    # Better explanation in ALE: the adverse consequences of extrapolation in PD plots
    # Ref: Visualizing the Effects of Predictor Variables in Black Box Supervised Learning Models


    feature_count = X.shape[1]

    # Select X to speed up calculation
    select_mask = weight != 0
    X = X[select_mask]
    weight = weight[select_mask]

    # Partial result of each features
    feature_list = []

    for feature_index in range(feature_count):
        pdp = partial_dependence(
            local_estimator,
            X,
            [feature_index],
            kind='both'
        )

        # Must get individual partial dependence to weight the result
        # Weight: Performance Weight. The point with more weight performance better in the model.
        # So average the partial performance according to the weight.
        individual = pdp['individual'][0]
        values = pdp['values'][0]
        weight_average = np.average(individual, axis=0, weights=weight)

        # TODO: Pack the result
        feature_list.append({
            'x': values,
            'pd': weight_average
        })

    return feature_list

# TODO: Add local ICE



import numpy as np
from numba import njit, prange
from numpy import ndarray
from scipy.sparse import csr_array


def second_order_neighbour(neighbour_matrix, neighbour_leave_out=None):
    """
    Calculate second-order neighbour matrix.
    Args:
        neighbour_matrix: First-order neighbour matrix.
        neighbour_leave_out: The subset of neighbours that should be considered as first-order neighbour. If None, use neighbour_matrix.


    Returns:

    """
    if neighbour_leave_out is None:
        neighbour_leave_out = neighbour_matrix

    if isinstance(neighbour_matrix, ndarray):
        return _second_order_neighbour_dense(neighbour_matrix, neighbour_leave_out)
    elif isinstance(neighbour_matrix, csr_array):
        indices_list = _second_order_neighbour_sparse(
            neighbour_matrix.indptr,
            neighbour_matrix.indices,
            neighbour_leave_out.indptr,
            neighbour_leave_out.indices,
        )

        # Generate the indptr and indices for the sparse matrix.
        indptr = np.zeros((len(indices_list) + 1,), dtype=np.int32)
        for i in range(len(indices_list)):
            indptr[i + 1] = indptr[i] + len(indices_list[i])

        indices = np.hstack(indices_list)

        return csr_array((np.ones_like(indices), indices, indptr))

    raise ValueError("neighbour_matrix should be np.ndarray or csr_array.")


@njit(parallel=True)
def _second_order_neighbour_sparse(
    indptr, indices, indptr_leave_out, indices_leave_out
):
    N = len(indptr) - 1
    # Manually create the list with specified length to avoid parallel Mutating error.
    indices_list = [np.empty(0, dtype=np.int64)] * N
    for row_index in prange(N):
        neighbour_indices = indices_leave_out[
            indptr_leave_out[row_index] : indptr_leave_out[row_index + 1]
        ]
        second_neighbour_indices_union = np.zeros((N,))
        for neighbour_index in neighbour_indices:
            second_neighbour_indices = indices[
                indptr[neighbour_index] : indptr[neighbour_index + 1]
            ]
            for second_neighbour_index in second_neighbour_indices:
                second_neighbour_indices_union[second_neighbour_index] = True

        second_neighbour_indices_union = np.nonzero(second_neighbour_indices_union)[0]
        indices_list[row_index] = second_neighbour_indices_union

    return indices_list


@njit(parallel=True)
def _second_order_neighbour_dense(neighbour_matrix, neighbour_leave_out):
    second_order_matrix = np.empty((neighbour_matrix.shape[1], neighbour_matrix.shape[1]), dtype=np.bool_)
    for i in prange(neighbour_matrix.shape[1]):
        second_order_matrix[i] = np.sum(
            neighbour_matrix[neighbour_leave_out[i]], axis=0
        )
    return second_order_matrix


def neighbour_shrink(weight_matrix, shrink_rate, return_weight_matrix=False, inplace=True):
    if not inplace:
        weight_matrix = weight_matrix.copy()

    if isinstance(weight_matrix, np.ndarray):
        weight_matrix = _neighbour_shrink(weight_matrix, shrink_rate)
        if return_weight_matrix:
            return weight_matrix
        else:
            return weight_matrix > 0

    elif isinstance(weight_matrix, csr_array):
        weight_matrix.data = _neighbour_shrink_sparse(weight_matrix.data, weight_matrix.indptr, shrink_rate)
        weight_matrix.eliminate_zeros()
        if return_weight_matrix:
            return weight_matrix
        else:
            return weight_matrix > 0

@njit(parallel=True)
def _neighbour_shrink(weight_matrix: np.ndarray, shrink_rate=0.5):
    for i in prange(weight_matrix.shape[0]):
        neighbour_indices = np.nonzero(weight_matrix[i])[0]
        positive_value = weight_matrix[i]
        positive_value = positive_value[neighbour_indices]
        shrink_value = np.quantile(positive_value, shrink_rate)
        positive_value[positive_value < shrink_value] = 0
        # TODO: Rename j
        for j in range(len(neighbour_indices)):
            weight_matrix[i, neighbour_indices[j]] = positive_value[j]
    return weight_matrix

@njit(parallel=True)
def _neighbour_shrink_sparse(weight_matrix_data, weight_matrix_indptr, shrink_rate=0.5):
    for i in prange(len(weight_matrix_indptr) - 1):
        positive_value = weight_matrix_data[
            weight_matrix_indptr[i] : weight_matrix_indptr[i + 1]
        ]
        shrink_value = np.quantile(positive_value, shrink_rate)
        positive_value[positive_value < shrink_value] = 0
        weight_matrix_data[
            weight_matrix_indptr[i] : weight_matrix_indptr[i + 1]
        ] = positive_value
    return weight_matrix_data


def sample_neighbour(weight_matrix, sample_rate, shrink_rate=None):
    """
    # TODO: More detailed description.

    Sample neighbour from weight matrix.
    Only the sampled neighbour will be used to fit the meta model.
    Therefore, the meta model will not be used for the sampled neighbour, but the out-of-sample neighbour.
    Args:
        weight_matrix:
        sample_rate:

    Returns:

    """

    # Do the shrink first.
    if shrink_rate is not None:
        neighbour_matrix = neighbour_shrink(weight_matrix, shrink_rate, inplace=False)
    else:
        neighbour_matrix = weight_matrix > 0

    # Do not sample itself.
    if isinstance(neighbour_matrix, np.ndarray):
        np.fill_diagonal(neighbour_matrix, False)
    elif isinstance(neighbour_matrix, csr_array):
        neighbour_matrix.setdiag(False)
        neighbour_matrix.eliminate_zeros()
    else:
        raise ValueError("weight_matrix should be np.ndarray or csr_array.")

    # Get the count to sample for each row.
    neighbour_count = np.sum(neighbour_matrix, axis=1)
    neighbour_count_sampled = np.ceil(neighbour_count * sample_rate).astype(int)
    neighbour_count_sampled[neighbour_count_sampled == 0] = 1
    neighbour_count_sampled[
        neighbour_count_sampled > neighbour_count
    ] = neighbour_count[neighbour_count_sampled > neighbour_count]

    neighbour_matrix_sampled = np.zeros(neighbour_matrix.shape, dtype=bool)

    # Set fixed random seed.
    np.random.seed(0)

    if isinstance(neighbour_matrix, np.ndarray):
        for i in range(neighbour_matrix.shape[0]):
            neighbour_matrix_sampled[
                i,
                np.random.choice(
                    # nonzero [0] for 1d array; [1] for 2d array.
                    np.nonzero(neighbour_matrix[i])[0],
                    neighbour_count_sampled[i],
                    replace=False,
                ),
            ] = True
    else:
        indices_list = []
        for i in range(neighbour_matrix.shape[0]):
            indices_list.append(
                # Sort the indices to make sure the structure of sparse matrix is correct.
                # But, really need to sort?
                np.sort(
                    # Leave out itself.
                    np.append(
                        np.random.choice(
                            neighbour_matrix.indices[
                                neighbour_matrix.indptr[i] : neighbour_matrix.indptr[
                                    i + 1
                                ]
                            ],
                            neighbour_count_sampled[i],
                            replace=False,
                        ),
                        i,
                    )
                )
            )

        indptr = np.zeros((len(indices_list) + 1,), dtype=np.int32)
        for i in range(len(indices_list)):
            indptr[i + 1] = indptr[i] + len(indices_list[i])

        indices = np.hstack(indices_list)
        neighbour_matrix_sampled = csr_array(
            (np.ones_like(indices), indices, indptr), dtype=bool
        )

    # Leave out itself.
    if isinstance(neighbour_matrix_sampled, np.ndarray):
        np.fill_diagonal(neighbour_matrix_sampled, True)

    return neighbour_matrix_sampled



from time import time as t

import numpy as np
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import RidgeCV
from sklearn.tree import DecisionTreeRegressor, ExtraTreeRegressor

from georegression.stacking_model import StackingWeightModel
from georegression.test.data import load_HP
from georegression.weight_model import WeightModel

# X, y_true, xy_vector, time = load_TOD()
# X, y_true, xy_vector, time = load_ESI()
X, y_true, xy_vector, time = load_HP()


def test_stacking():
    local_estimator = DecisionTreeRegressor(splitter="random", max_depth=2)
    distance_measure = "euclidean"
    kernel_type = "bisquare"
    distance_ratio = None
    bandwidth = None
    neighbour_count = 0.01

    model = StackingWeightModel(
        local_estimator,
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
    )

    model.fit(X, y_true, [xy_vector, time])
    print(f"{model.llocv_score_}, {model.llocv_stacking_}")


def test_alpha():
    local_estimator = DecisionTreeRegressor(splitter="random", max_depth=2)
    distance_measure = "euclidean"
    kernel_type = "bisquare"
    distance_ratio = None
    bandwidth = None
    neighbour_count = 0.01

    model = StackingWeightModel(
        local_estimator,
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
        alpha=10,
    )

    model.fit(X, y_true, [xy_vector, time])
    print(f"{model.llocv_score_}, {model.llocv_stacking_}")

    for local_estimator in model.local_estimator_list:
        print(local_estimator.meta_estimator.coef_)
        break

    # For alpha=0.1, stacking_score = 0.5750569627981988
    """
    Coefficients of first stacking estimator:
    [-0.46379083 -0.38453714  0.39963185  0.01484807  0.16410479 -0.59694787
      0.21276714  0.11330034  0.29212005 -0.20581994  0.07942222  0.92542167
      0.44300962  0.26067723  0.03980381 -0.32809317  0.17886772  0.26176183
      0.31227637  0.12423833  0.23946592]
    """

    # For alpha=10, stacking_score = 0.9403979818713938
    """
    Coefficients of first stacking estimator:
    [ 0.07789433 -0.03072463  0.18275214 -0.00193438  0.05766076 -0.00123777
      0.13473063  0.1755927   0.0568057   0.0234573   0.14681941  0.03860493
     -0.06496593  0.1208457   0.06717717  0.0523331   0.0167307   0.14635798
     -0.03296376 -0.04416956  0.26379955]
    """


def test_estimator_sample():
    local_estimator = DecisionTreeRegressor(splitter="random", max_depth=2)
    distance_measure = "euclidean"
    kernel_type = "bisquare"
    distance_ratio = None
    bandwidth = None
    neighbour_count = 0.1

    model = StackingWeightModel(
        local_estimator,
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
        estimator_sample_rate=0.1,
    )

    model.fit(X, y_true, [xy_vector, time])
    print(f"{model.llocv_score_}, {model.llocv_stacking_}")

    model = StackingWeightModel(
        local_estimator,
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
        estimator_sample_rate=0.5,
    )
    model.fit(X, y_true, [xy_vector, time])
    print(f"{model.llocv_score_}, {model.llocv_stacking_}")

    model = StackingWeightModel(
        local_estimator,
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
        estimator_sample_rate=None,
    )
    model.fit(X, y_true, [xy_vector, time])
    print(f"{model.llocv_score_}, {model.llocv_stacking_}")


def test_performance():
    distance_measure = "euclidean"
    kernel_type = "bisquare"
    distance_ratio = None
    bandwidth = None
    neighbour_count = 0.1

    estimator = StackingWeightModel(
        ExtraTreeRegressor(splitter="random", max_depth=1),
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
        neighbour_leave_out_rate=0.1,
        # estimator_sample_rate=0.1,
    )

    t1 = t()
    estimator.fit(X, y_true, [xy_vector, time])
    t2 = t()
    print(t2 - t1, estimator.llocv_score_, estimator.llocv_stacking_)
    # neighbour_count = 0.1 neighbour_leave_out_rate=0.1 17.498192310333252 0.7670304978651743 0.8134413795992002

    estimator = WeightModel(
        RandomForestRegressor(n_estimators=50),
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
    )
    t2 = t()
    estimator.fit(X, y_true, [xy_vector, time])
    t3 = t()

    print(t3 - t2, estimator.llocv_score_)
    # neighbour_count = 0.1 n_estimators=50 34.99542546272278 0.8096408618045396

    estimator = WeightModel(
        RidgeCV(),
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
    )
    estimator.fit(X, y_true, [xy_vector, time])
    t4 = t()
    # neighbour_count = 0.1 5.488587141036987 0.7706704632683226

    print(t4 - t3, estimator.llocv_score_)


def test_stacking_not_leaking():
    # local_estimator = DecisionTreeRegressor(splitter="random", max_depth=2)
    # local_estimator = DecisionTreeRegressor()
    local_estimator = RandomForestRegressor(n_estimators=50)
    # local_estimator = RidgeCV()
    distance_measure = "euclidean"
    kernel_type = "bisquare"
    distance_ratio = None
    bandwidth = None

    def fit_wrapper():
        estimator = StackingWeightModel(
            local_estimator,
            distance_measure,
            kernel_type,
            distance_ratio,
            bandwidth,
            neighbour_count,
            neighbour_leave_out_rate=leave_out_rate,
        )

        t1 = t()
        estimator.fit(X, y_true, [xy_vector, time])
        t2 = t()
        print("neighbour_count =", neighbour_count, "leave_out_rate =", leave_out_rate)
        print(
            "time =",
            t2 - t1,
            "llocv_score =",
            estimator.llocv_score_,
            "llocv_stacking =",
            estimator.llocv_stacking_,
        )

    neighbour_count = 0.1
    leave_out_rate = 0.1

    fit_wrapper()

    estimator = WeightModel(
        RandomForestRegressor(n_estimators=50),
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
    )
    t2 = t()
    estimator.fit(X, y_true, [xy_vector, time])
    t3 = t()
    print(t3 - t2, estimator.llocv_score_)

    for depth in range(1, 31, 2):
        # local_estimator = DecisionTreeRegressor(max_depth=depth)
        # fit_wrapper()
        pass

    for neighbour_count in np.arange(0.05, 0.35, 0.05):
        for leave_out_rate in np.arange(0.05, 0.35, 0.05):
            # fit_wrapper()
            pass


if __name__ == "__main__":
    # test_stacking()
    # test_alpha()
    # test_estimator_sample()
    test_performance()
    # test_stacking_not_leaking()



from scipy.sparse import csr_array
from sklearn.linear_model import LinearRegression, Ridge
from sklearn.tree import DecisionTreeRegressor, ExtraTreeRegressor
from sklearn.metrics import r2_score

from georegression.stacking_model import StackingWeightModel
from georegression.test.data import load_HP
from georegression.weight_matrix import weight_matrix_from_points
from georegression.weight_model import WeightModel

from time import time as t

X, y_true, xy_vector, time = load_HP()


def test_compatibility():
    weight_matrix = weight_matrix_from_points(
        [xy_vector, time], [xy_vector, time], "euclidean", "bisquare", None, None, 0.1, None
    )

    # Normal case
    t1 = t()
    estimator = WeightModel(
        ExtraTreeRegressor(max_depth=1, splitter="random"),
        "euclidean",
        "bisquare",
        neighbour_count=0.1
    )
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    t2 = t()
    print(t2 - t1)
    print(estimator.llocv_score_)

    # Sparse case
    weight_matrix = csr_array(weight_matrix)

    t1 = t()
    estimator = WeightModel(
        ExtraTreeRegressor(max_depth=1, splitter="random"),
        "euclidean",
        "bisquare",
        neighbour_count=0.1
    )
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    t2 = t()
    print(t2 - t1)
    print(estimator.llocv_score_)


def test_stacking_compatibility():
    weight_matrix = weight_matrix_from_points(
        [xy_vector, time],
        [xy_vector, time],
        "euclidean",
        "bisquare", None, None,
        0.1, None
    )

    # Normal case
    estimator = StackingWeightModel(
        ExtraTreeRegressor(max_depth=1, splitter="random"),
        neighbour_leave_out_rate=0.1,
        use_numba=False
    )
    # estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    print(estimator.llocv_score_)
    print(estimator.llocv_stacking_)

    # Sparse case
    weight_matrix = csr_array(weight_matrix)

    estimator = StackingWeightModel(
        ExtraTreeRegressor(max_depth=1, splitter="random"),
        neighbour_leave_out_rate=0.1,
        use_numba=False
    )
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    print(estimator.llocv_score_)
    print(estimator.llocv_stacking_)

    print()

if __name__ == '__main__':
    # test_compatibility()
    test_stacking_compatibility()



from hyperopt import hp, fmin, tpe, Trials


# 1. Setting up the Custom Search Space
def custom_prior(name):
    def sample_and_clip():
        # 80% chance of choosing the normal distribution
        chosen_distr = hp.choice(name + "_choice", [0, 1])
        if chosen_distr == 0:
            val = hp.normal(name + "_normal", 0.25, 0.05)
            return max(0, min(1, val))  # Clip the value to the [0, 1] range
        else:
            return hp.uniform(name + "_uniform", 0, 1)

    return sample_and_clip()


space = {"param_name": custom_prior("param_name")}


# 2. Defining the Objective Function
def objective(params):
    # Clipping the parameter value to ensure it's in the [0, 1] range
    params["param_name"] = max(0, min(1, params["param_name"]))

    # Hypothetical objective function (replace with your actual function)
    loss = (params["param_name"] - 0.3) ** 2  # Minimized when param_name is 0.3

    return loss


# 3. Optimization
trials = Trials()
best = fmin(fn=objective, space=space, algo=tpe.suggest, max_evals=100, trials=trials)

print("Best hyperparameter value:", best)



def get_toy_model():
    from georegression.test.data import load_HP
    X, y, xy_vector, time = load_HP()
    from georegression.weight_model import WeightModel
    from sklearn.linear_model import LinearRegression
    model = WeightModel(
        LinearRegression(),
        distance_measure='euclidean',
        kernel_type='bisquare',
        neighbour_count=0.5,
        cache_data=True, cache_estimator=True
    )
    model.fit(X[:100, :10], y[:100], [xy_vector[:100], time[:100]])
    model.partial_dependence()

    return model



import hyperopt
from hyperopt import fmin, tpe, hp, Trials
import numpy as np


# Define the search space
def custom_prior(name):
    return hp.choice(
        name,
        [hp.normal(name + "_normal", 0.25, 0.05), hp.uniform(name + "_uniform", 0, 1)],
    )


space = {"param_name": custom_prior("param_name")}


# Define the objective function
def objective(params):

    if params["param_name"] < 0.25:
        return {"status": hyperopt.STATUS_FAIL}

    # Ensure the parameter is within [0, 1]
    params["param_name"] = max(0, min(1, params["param_name"]))

    # Fictional objective function: (x-0.25)^2, the minimum is at x=0.25
    loss = (params["param_name"] - 0.25) ** 2

    return {"loss": loss, "status": hyperopt.STATUS_OK}


# Optimization
trials = Trials()
best = fmin(fn=objective, space=space, algo=tpe.suggest, max_evals=1000, trials=trials)

print(f"Best hyperparameter value: {best['param_name']}")



from scipy.sparse import csr_array
from sklearn.linear_model import LinearRegression, Ridge
from sklearn.tree import DecisionTreeRegressor, ExtraTreeRegressor
from sklearn.metrics import r2_score

from georegression.stacking_model import StackingWeightModel
from georegression.test.data import load_HP
from georegression.weight_matrix import weight_matrix_from_points
from georegression.weight_model import WeightModel

from time import time as t

X, y_true, xy_vector, time = load_HP()

def test_neighbour_leave_out_shrink_rate():
    weight_matrix = weight_matrix_from_points(
        [xy_vector, time], [xy_vector, time], "euclidean", "bisquare", None, None, 0.1, None
    )

    estimator = StackingWeightModel(
        ExtraTreeRegressor(max_depth=1, splitter="random"),
        neighbour_leave_out_rate=0.1,
        use_numba=False,
        neighbour_leave_out_shrink_rate=0.3,
    )
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    print(estimator.llocv_score_)
    print(estimator.llocv_stacking_)

    weight_matrix = csr_array(weight_matrix)
    estimator.use_numba = True
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    print(estimator.llocv_score_)
    print(estimator.llocv_stacking_)


def test_meta_fitting_shrink_rate():
    weight_matrix = weight_matrix_from_points(
        [xy_vector, time], [xy_vector, time], "euclidean", "bisquare", None, None, 0.1, None
    )
    
    estimator = StackingWeightModel(
        ExtraTreeRegressor(max_depth=1, splitter="random"),
        neighbour_leave_out_rate=0.1,
        use_numba=False,
        meta_fitting_shrink_rate=0.9,
    )
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    print(estimator.llocv_score_)
    print(estimator.llocv_stacking_)
    
    weight_matrix = csr_array(weight_matrix)
    estimator.use_numba = True
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    print(estimator.llocv_score_)
    print(estimator.llocv_stacking_)



if __name__ == '__main__':
    # test_neighbour_leave_out_shrink_rate()
    test_meta_fitting_shrink_rate()


import time

import numpy as np
from joblib.parallel import Parallel, delayed
from sklearn.ensemble import RandomForestRegressor
from sklearn.tree import DecisionTreeRegressor


def test_tree_efficiency():
    """
    对于某局部模型，给定局部数据X
    Stacking：训练单个简单的树，对X个邻接的X进行预测。
    训练时间train*data(X)，预测时间predict*data(X)*number(X)。实际中预测的X有重复，应小于这个值。
    线性Stacking时间Inverse(X*X) X个样本，X个邻接提供的X个特征。

    RandomForest：训练n个完全树，进行平均。
    训练时间train*data(X)*number(n)

    Returns:

    """
    neighbour_count = 5000
    feature_count = 30
    X = np.random.random((neighbour_count, feature_count))
    X_bulk = np.concatenate([X] * neighbour_count, axis=0)
    y = np.random.random(neighbour_count)

    # Wamp-up

    tree = DecisionTreeRegressor()
    tree.fit(X, y)
    tree.predict(X)

    forest = RandomForestRegressor()
    forest.fit(X, y)

    t1 = time.time()

    tree = DecisionTreeRegressor(max_depth=2, splitter='random')
    tree.fit(X, y)

    t2 = time.time()

    # Predict in atomic. But not completely atomic.
    for i in range(neighbour_count):
        tree.predict(X)

    t3 = time.time()

    # Bulk is better for moderate amount of data. (100~500)^2. Degrade when come to 1000^2.
    # https://scikit-learn.org/0.15/modules/computational_performance.html
    # https://scikit-learn.org/0.15/auto_examples/applications/plot_prediction_latency.html#example-applications-plot-prediction-latency-py
    tree.predict(X_bulk)

    t4 = time.time()

    forest = RandomForestRegressor(n_estimators=100, n_jobs=-1)
    forest.fit(X, y)

    t5 = time.time()

    print()
    print(t2 - t1, t3 - t2, t4 - t3, t5 - t4)
    # neighbour = 100 0.0 0.00400090217590332 0.0009999275207519531 0.09102034568786621
    # neighbour = 500 0.0 0.02702188491821289 0.016988277435302734 0.12402749061584473
    # neighbour = 1000 0.0009999275207519531 0.06401467323303223 0.06501483917236328 0.1708080768585205
    # neighbour = 2000 0.0010001659393310547 0.18304109573364258 0.2760617733001709 0.3486180305480957
    # neighbour = 5000 0.003000497817993164 0.8315958976745605 3.4696438312530518 1.164928913116455


def fit_estimator(estimator, X, y):
    estimator.fit(X, y)
    return estimator


def estimators_predict(estimator_list, X):
    result_list = []
    for estimator in estimator_list:
        result_list.append(
            estimator.predict_by_weight(X)
        )
    return result_list


def estimators_parallel_predict(estimator_list, X):
    return Parallel(-1)(
        delayed(estimator_predict)(
            estimator, X
        )
        for estimator in estimator_list
    )


def estimator_predict(estimator, X):
    return estimator.predict_by_weight(X)


def estimator_predict_n_time(estimator, X, times):
    predict_result = []
    for i in range(times):
        predict_result.append(
            estimator.predict_by_weight(X)
        )
    return predict_result


def test_in_joblib():
    # 邻接数量
    neighbour_count = 100
    # 局部模型数量
    estimator_count = 100

    feature_count = 30
    X = np.random.random((neighbour_count, feature_count))
    X_bulk = np.concatenate([X] * neighbour_count, axis=0)
    y = np.random.random(neighbour_count)

    parallel = Parallel(-1)

    # Warm-up
    parallel([
        delayed(fit_estimator)(
            DecisionTreeRegressor(), X, y
        )
        for i in range(64)
    ])
    parallel([
        delayed(fit_estimator)(
            RandomForestRegressor(), X, y
        )
        for i in range(64)
    ])

    print('Tree fitting starts.')
    t1 = time.time()

    tree_list = parallel([
        delayed(fit_estimator)(
            DecisionTreeRegressor(max_depth=2, splitter='random'), X, y
        )
        for i in range(estimator_count)
    ])

    print('Prediction Starts.')
    t2 = time.time()

    # Really slow. Resource not fully used.

    # Only get the theoretical time when neighbour count equals estimator count.
    # i.e. tree list equals neighbour estimator list.
    tree_result = parallel([
        delayed(estimators_predict)(
            tree_list, X
        )
        for i in range(estimator_count)
    ])

    t3 = time.time()

    tree_result = parallel(
        delayed(estimator_predict_n_time)(
            estimator, X, neighbour_count
        )
        for estimator in tree_list
    )

    t4 = time.time()

    for estimator in tree_list:
        for i in range(neighbour_count):
            estimator.predict_by_weight(X)

    t5 = time.time()

    for estimator in tree_list:
        estimator.predict_by_weight(X_bulk)

    t6 = time.time()

    tree_result = parallel(
        delayed(estimator_predict)(
            estimator, X_bulk
        )
        for estimator in tree_list
    )

    t7 = time.time()

    split_num = 6
    tree_result = parallel(
        delayed(estimators_predict)(
            [tree_list[i] for i in split_index], X_bulk
        )
        for split_index in np.array_split(np.arange(estimator_count), split_num)
    )

    print('Prediction Ends.')
    t8 = time.time()

    forest_list = parallel([
        delayed(fit_estimator)(
            RandomForestRegressor(n_estimators=100, n_jobs=-1), X, y
        )
        for i in range(estimator_count)
    ])

    t9 = time.time()

    print()
    print(t2 - t1)
    print(t3 - t2, t4 - t3, t5 - t4, t6 - t5, t7 - t6, t8 - t7)
    print(t9 - t8)
    # neighbour = estimator = 100
    # 0.03500962257385254
    # 0.65926194190979 0.09602212905883789 0.3746037483215332 0.05401206016540527 0.17014646530151367 0.05001258850097656
    # 1.7216594219207764


if __name__ == '__main__':
    test_in_joblib()



import os
from functools import partial

import matplotlib.pyplot as plt
from sklearn.ensemble import RandomForestRegressor

from georegression.local_ale import weighted_ale
from georegression.simulation.simulation_utils import *
from georegression.visualize.ale import plot_ale
from georegression.weight_model import WeightModel

# Font family
plt.rcParams["font.family"] = "Times New Roman"
plt.rcParams["font.size"] = 18
plt.rcParams["axes.labelsize"] = 18
plt.rcParams['font.weight'] = 'bold'
plt.rcParams['xtick.labelsize'] = 15
plt.rcParams['ytick.labelsize'] = 15

def coef_manual_gau():
    coef_radial = radial_coefficient(np.array([0, 0]), 1 / np.sqrt(200))
    coef_dir = directional_coefficient(np.array([1, 1]))

    coef_gau_1 = gaussian_coefficient(np.array([-5, 5]), [[3, 4], [4, 8]], amplitude=-1)
    coef_gau_2 = gaussian_coefficient(np.array([-2, -5]), 5, amplitude=2)
    coef_gau_3 = gaussian_coefficient(np.array([8, 3]), 10, amplitude=-1.5)
    coef_gau_4 = gaussian_coefficient(
        np.array([2, 8]), [[3, 0], [0, 15]], amplitude=0.8
    )
    coef_gau_5 = gaussian_coefficient(np.array([5, -10]), 1, amplitude=1)
    coef_gau_6 = gaussian_coefficient(np.array([-10, -10]), 15, amplitude=1.5)
    coef_gau_6 = gaussian_coefficient(np.array([-11, 0]), 5, amplitude=2)
    coef_gau_6 = gaussian_coefficient(np.array([-11, 0]), 5, amplitude=2)
    coef_gau = coefficient_wrapper(
        np.sum, coef_gau_1, coef_gau_2, coef_gau_3, coef_gau_4, coef_gau_5, coef_gau_6
    )

    # coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_dir, coef_gau)
    coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_gau)

    return coef_sum

def coef_auto_gau_strong():
    coef_radial = radial_coefficient(np.array([0, 0]), 1 / np.sqrt(200))
    coef_dir = directional_coefficient(np.array([1, 1]))

    gau_coef_list = []
    for _ in range(1000):
        # Randomly generate the parameters for gaussian coefficient
        center = np.random.uniform(-10, 10, 2)
        amplitude = np.random.uniform(1, 2)
        sign = np.random.choice([-1, 1])
        amplitude *= sign
        sigma1 = np.random.uniform(0.2, 1)
        sigma2 = np.random.uniform(0.2, 1)
        cov = np.random.uniform(-np.sqrt(sigma1 * sigma2), np.sqrt(sigma1 * sigma2))
        sigma = np.array([[sigma1, cov], [cov, sigma2]])

        coef_gau = gaussian_coefficient(center, sigma, amplitude=amplitude)
        gau_coef_list.append(coef_gau)

    coef_gau = coefficient_wrapper(np.sum, *gau_coef_list)
    coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_dir, coef_gau)

    return coef_sum

def f_interact(X, C, points):
    return interaction_function(C[0])(X[:, 0], X[:, 1], points) + 0


f = f_interact
coef_func = coef_manual_gau
x2_coef = coefficient_wrapper(partial(np.multiply, 3), coef_func())


def generate_sample(random_seed=1):
    np.random.seed(random_seed)

    count = 5000
    points = sample_points(count, bounds=[[-10, 10], [-10, 10]])
    x1 = sample_x(count, mean=coef_func(), bounds=(-1, 1), points=points)
    x2_coef = coefficient_wrapper(partial(np.multiply, 3), coef_func())
    x2 = sample_x(count, mean=x2_coef, bounds=(-2, 2), points=points)

    if isinstance(coef_func, list):
        coefficients = [func() for func in coef_func]
    else:
        coefficients = [coef_func()]

    X = np.stack((x1, x2), axis=-1)
    y = f(X, coefficients, points)

    return X, y, points, f, coefficients


def draw_graph():
    X, y, points, f, coef = generate_sample()
    X_plus = np.concatenate([X, points], axis=1)
    distance_measure = "euclidean"
    kernel_type = "bisquare"
    neighbour_count = 0.05

    # local_estimator = DecisionTreeRegressor(splitter="random", max_depth=1)
    # local_estimator = DecisionTreeRegressor(splitter="random", max_depth=2)
    # model = StackingWeightModel(
    #     local_estimator,
    #     distance_measure,
    #     kernel_type,
    #     neighbour_count=neighbour_count,
    #     neighbour_leave_out_rate=0.25,
    #     cache_data=True,
    #     cache_estimator=True,
    # )
    # model.fit(X, y, [points])
    # print('Stacking:', model.llocv_score_, model.llocv_stacking_)

    model = WeightModel(
        RandomForestRegressor(n_estimators=50),
        distance_measure,
        kernel_type,
        neighbour_count=neighbour_count,
        cache_data=True,
        cache_estimator=True,
    )
    model.fit(X, y, [points])
    print("GRF:", model.llocv_score_)

    feature_index = 0

    for local_index in range(model.N):
        estimator = model.local_estimator_list[local_index]
        neighbour_mask = model.neighbour_matrix_[local_index]
        neighbour_weight = model.weight_matrix_[local_index][neighbour_mask]
        X_local = model.X[neighbour_mask]
        ale_result = weighted_ale(
            X_local, feature_index, estimator.predict, neighbour_weight
        )

        fval, ale = ale_result

        x_neighbour = X[model.neighbour_matrix_[local_index], feature_index]
        y_neighbour = y[model.neighbour_matrix_[local_index]]
        weight_neighbour = model.weight_matrix_[
            local_index, model.neighbour_matrix_[local_index]
        ]

        # show_function_at_point(f, coef, points[local_index], ax=ax)
        # Get the true marginal effect for function f = x1 * x2.
        x_gird = np.linspace(np.min(x_neighbour), np.max(x_neighbour), 1000)
        x1 = X[local_index, 1]
        x1 = np.tile(x1, 1000)

        # X_grid = np.stack([x_gird, x1], axis=-1)
        # y_grid = f(X_grid, coef, points[local_index])

        from georegression.simulation.simulation import x2_coef

        beta = coef[0](points[local_index])
        x2_average = x2_coef(points[local_index])
        y_grid = (
            beta
            * 0.5
            * ((x2_average + 2) ** 2 - (x2_average - 2) ** 2)
            * (1 / 4)
            * x_gird
        )

        x1_base = np.empty(500)
        x1_base[:] = np.min(x_neighbour)
        x2_base = np.random.uniform(x2_average - 2, x2_average + 2, 500)
        base_value_real = estimator.predict(
            np.stack([x1_base, x2_base], axis=-1)
        ).mean()

        diff = ale[0] - base_value_real
        ale = ale - diff

        fig = plot_ale(fval, ale, x_neighbour)
        fig.set_size_inches(10, 6)
        ax1 = fig.get_axes()[0]
        ax2 = fig.get_axes()[1]

        ax1.set_xlabel("Feature value", fontweight='bold')
        ax1.set_ylabel("Function value", fontweight='bold')
        ax2.set_ylabel('Density', fontweight='bold')

        scatter = ax1.scatter(x_neighbour, y_neighbour, c=weight_neighbour)
        ax1.scatter(
            X[local_index, feature_index], y[local_index], c="red", label="Local point"
        )
        cbar = fig.colorbar(scatter, ax=ax1, label="Weight", pad=0.1)
        cbar.set_label('Weight', weight='bold')
        cbar.ax.tick_params(labelsize=15)

        ax1.plot(x_gird, y_grid, label="True value")

        # Neighbor ALE, which only consider the neighbor points but not weight is considered
        # ale_result = weighted_ale(
        #     X_local, feature_index, estimator.predict, np.ones(X_local.shape[0])
        # )
        # fval, ale = ale_result
        # diff = ale[0] - base_value_real
        # ale = ale - diff
        # ax.plot(fval, ale, label="Neighbour ALE")

        # Add non-weighted ALE plot
        # Select the X that is in the value range of x_neighbour
        x_global_ale = X[
            (X[:, feature_index] >= np.min(x_neighbour))
            & (X[:, feature_index] <= np.max(x_neighbour))
        ]
        ale_result = weighted_ale(
            x_global_ale,
            feature_index,
            estimator.predict,
            np.ones(x_global_ale.shape[0]),
        )
        fval, ale = ale_result

        x1_base = np.empty(500)
        x1_base[:] = np.min(x_neighbour)
        x2_base = np.random.choice(X[:, 1], 500)
        base_value_real = estimator.predict(
            np.stack([x1_base, x2_base], axis=-1)
        ).mean()

        diff = ale[0] - base_value_real
        ale = ale - diff

        ax1.plot(fval, ale, label="ALE")

        # ALE True value
        # y_grid_ale = (x2_coef(points[local_index])* x_gird)
        # ax.plot(x_gird, y_grid_ale, label="True ALE")

        # Add legend
        handles, labels = ax1.get_legend_handles_labels()
        handles.append(scatter)
        labels.append("Weight")
        ax1.legend(handles, labels, fontsize=15)

        folder_name = "Plot/LocalAle_BigFont"
        os.makedirs(folder_name, exist_ok=True)
        plt.savefig(f"{folder_name}/{local_index}.png", dpi=300)
        plt.close()
        # plt.show(block=True)


if __name__ == "__main__":
    draw_graph()



from sklearn.tree import ExtraTreeRegressor

from georegression.stacking_model import StackingWeightModel
from georegression.test.data import load_HP

X, y_true, xy_vector, time = load_HP()


def test_logging():
    distance_measure = "euclidean"
    kernel_type = "bisquare"
    distance_ratio = None
    bandwidth = None
    neighbour_count = 0.1

    estimator = StackingWeightModel(
        ExtraTreeRegressor(splitter="random", max_depth=1),
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
        neighbour_leave_out_rate=0.1,
    )

    estimator.fit(X, y_true, [xy_vector, time])


if __name__ == "__main__":
    test_logging()



from sklearn.metrics import *

from georegression.test.data import load_HP

(X, y_true, xy_vector, time) = load_HP()


def test_rf():
    from sklearn.ensemble import RandomForestRegressor

    print('RandomForest Model')
    estimator = RandomForestRegressor(n_estimators=1000, n_jobs=-1, oob_score=True)
    estimator.fit(X, y_true)
    print(estimator.oob_score_)

    MAPE = mean_absolute_percentage_error(y_true, estimator.oob_prediction_)
    print(MAPE)

    return estimator.predict(estimator.oob_prediction_)




from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor
from georegression.test.data import load_HP
from georegression.weight_model import WeightModel
from georegression.visualize.ale import plot_ale

X, y, xy_vector, time = load_HP()
X = X[:200]
y = y[:200]
xy_vector = xy_vector[:200]
time = time[:200]

def test_ale():
    global X

    model = WeightModel(
        # LinearRegression(),
        RandomForestRegressor(n_estimators=10),
        distance_measure='euclidean',
        kernel_type='bisquare',
        neighbour_count=0.1,
        cache_data=True, cache_estimator=True, n_jobs=1
    )

    X = X[:, -5:]
    model.fit(X, y, [xy_vector, time])
    for i in range(5):
        feature_index = i
        fval, ale = model.global_ALE(feature_index)
        plot_ale(fval, ale, X[:, feature_index])

    print()

if __name__ == '__main__':
    test_ale()



from sklearn.linear_model import LinearRegression
from sklearn.metrics import r2_score

from georegression.test.data import load_HP
from georegression.weight_model import WeightModel

X, y, xy_vector, time = load_HP()


def test_predict():
    model = WeightModel(
        LinearRegression(),
        distance_measure='euclidean',
        kernel_type='bisquare',
        neighbour_count=0.1,

        cache_data=True, cache_estimator=True
    )

    model.fit(X, y, [xy_vector, time])
    prediction_by_weight = model.predict_by_weight(X[:10], [xy_vector[:10], time[:10]])
    prediction_by_fit = model.predict_by_fit(X[:10], [xy_vector[:10], time[:10]])

    # Prediction outperform the local_predict because of the data leak.
    print(
        r2_score(y[:10], model.local_predict_[:10]),
        r2_score(y[:10], prediction_by_weight),
        r2_score(y[:10], prediction_by_fit),
    )


if __name__ == '__main__':
    test_predict()



# TODO: https://jaykmody.com/blog/distance-matrices-with-numpy/
# TODO: https://stackoverflow.com/questions/22720864/efficiently-calculating-a-euclidean-distance-matrix-using-numpy
# TODO: Ref to https://github.com/talboger/fastdist and https://github.com/numba/numba-scipy/issues/38#issuecomment-623569703 to speed up by parallel computing
from pathlib import Path


import numpy as np
from numba import njit
from scipy.spatial.distance import pdist, cdist


def _distance_matrices(
    source_coords: list[np.ndarray], target_coords=None, metrics='euclidean',
        use_dask=False, cache_sort=False, args=None, **kwargs
):
    """
    Check the validatiton of the parameter list.
    If single value is provided, convert it to a list with length equal to the dimension of the vector list.
    Then, call the distance_matrix function for each dimension.

    Args:
        source_coords:
        target_coords:
        metrics:
        use_dask:
        args:

    Returns:

    """

    dimension = len(source_coords)

    # Check equal length of source and target coordinates
    if target_coords is not None:
        if dimension != len(target_coords):
            raise Exception("Source and target coordinate length not match")
    else:
        target_coords = [None] * dimension

    # Check whether the input parameters are lists, and if not, convert them to lists with length equal to the dimension of the vector list.
    if not isinstance(metrics, list):
        metrics = [metrics] * dimension

    if args is None:
        args = [kwargs] * dimension

    return [
        _distance_matrix(
            source_coords[dim],
            target_coords[dim],
            metrics[dim],
            use_dask,
            cache_sort,
            **args[dim],
        )
        for dim in range(dimension)
    ]


def _distance_matrix(source_coord, target_coord, metric, use_dask, cache_sort, **kwargs):
    # Check equal dimension of source and target coordinates
    if target_coord is not None:
        if source_coord.shape[1] != target_coord.shape[1]:
            raise Exception("Source and target coordinate dimension not match")

    dimension = source_coord.shape[1]

    if use_dask:
        import dask.array as da
        import dask_distance

        filepath = kwargs.get("filepath", "distance_matrix.zarr")

        if not filepath.endswith(".zarr"):
            filepath = filepath + ".zarr"

        if (kwargs.get("filepath", None) is not None) and (not kwargs.get("overwrite", False)):
            if Path(filepath).exists():
                return da.from_zarr(filepath)

        if target_coord is None:
            # TODO: Size error even after the rechunk.
            distance_matrix = dask_distance.pdist(source_coord, metric=metric)
        else:
            distance_matrix = dask_distance.cdist(source_coord, target_coord, metric=metric)

        distance_matrix = distance_matrix.rechunk({0: "auto", 1: -1})
        distance_matrix.to_zarr(filepath, overwrite=True)

        if cache_sort:
            distance_matrix_sorted = distance_matrix.map_blocks(np.sort)
            distance_matrix_sorted.to_zarr(filepath.replace(".zarr", "_sorted.zarr"), overwrite=True)

            return distance_matrix, distance_matrix_sorted

        return distance_matrix

    else:
        if metric == "great-circle":
            if dimension != 2:
                raise Exception("Great-circle distance only applicable to 2D coordinates")
            return np.array(
                [great_circle_distance(coord, target_coord) for coord in source_coord]
            ).astype(np.float32)

        if target_coord is None:
            return pdist(source_coord.astype(np.float32), metric, **kwargs).astype(np.float32)
        else:
            return cdist(
                source_coord.astype(np.float32),
                target_coord.astype(np.float32),
                metric,
                **kwargs,
            ).astype(np.float32)


@njit
def great_circle_distance(one_lonlat, many_lonlat):
    """
    Compute great-circle distance using Haversine algorithm.
    """

    lon_diff = np.radians(many_lonlat[:, 0] - one_lonlat[0])
    lat_diff = np.radians(many_lonlat[:, 1] - one_lonlat[1])
    lat_one = np.radians(one_lonlat[1])
    lat_many = np.radians(many_lonlat[:, 1])

    a = (
        np.sin(lat_diff / 2) ** 2
        + np.cos(lat_many) * np.cos(lat_one) * np.sin(lon_diff / 2) ** 2
    )
    c = 2 * np.arcsin(np.sqrt(a))
    R = 6371.0

    return R * c



import matplotlib.pyplot as plt


import numpy as np
from matplotlib import transforms
from sklearn.neighbors import KernelDensity


def plot_ale(fvals, ale, x):
    fig, ax1 = plt.subplots(figsize=(12, 6))

    ax1.plot(fvals, ale, zorder=2, label='STALE')
    ax1.grid(True, which='both', linestyle='--', linewidth=0.5)

    ax1.set_xlabel('Feature value')
    ax1.set_ylabel('ALE')

    # Density distribution
    ax2 = ax1.twinx()
    ax2.hist(x, bins=10, density=False, alpha=0.3, color='gray', zorder=1, label='Density')
    ax2.grid(False)

    ax2.set_ylabel('Density')

    return fig



import logging

from slab_utils.quick_logger import logger

# Change the format for file output
fileHandler = [handler for handler in logger.handlers if isinstance(handler, logging.FileHandler)][0]
formatter4File = logging.Formatter(
    '[%(levelname)s] - %(asctime)s:\n%(message)s',
    '%Y-%m-%d %H:%M:%S')
fileHandler.setFormatter(formatter4File)

# set LOKY_PICKLER=pickle using os
# os.environ['LOKY_PICKLER'] = 'pickle'

import matplotlib

matplotlib.use('TKAgg')



import time

import numpy as np
from sklearn.linear_model import LinearRegression, Ridge, RidgeCV


def test_regression():
    neighbour_count = 500
    X = np.random.random((neighbour_count, neighbour_count))
    y = np.random.random(neighbour_count)
    estimator_count = 100

    t1 = time.time()

    for i in range(estimator_count):
        e = LinearRegression()
        e.fit(X, y)

    t2 = time.time()

    for i in range(estimator_count):
        e = Ridge()
        e.fit(X, y)

    t3 = time.time()

    for i in range(estimator_count):
        e = RidgeCV()
        e.fit(X, y)

    t4 = time.time()

    print()
    print(t2 - t1, t3 - t2, t4 - t3)
    # 2.859616994857788 0.6065783500671387 3.498260974884033
    # neighbour_count = 500 estimator_count = 100 AMD-7950X 10.886837720870972 0.49491024017333984 6.453542470932007


def test_solver():
    neighbour_count = 1000
    X = np.random.random((neighbour_count, neighbour_count))
    y = np.random.random(neighbour_count)
    estimator_count = 100

    t1 = time.time()

    for i in range(estimator_count):
        e = Ridge(10)
        e.fit(X, y)

    t2 = time.time()

    for i in range(estimator_count):
        e = Ridge(10, solver="svd")
        e.fit(X, y)

    t3 = time.time()

    for i in range(estimator_count):
        e = Ridge(10, solver="cholesky")
        e.fit(X, y)

    t4 = time.time()

    for i in range(estimator_count):
        e = Ridge(10, solver="sparse_cg")
        e.fit(X, y)

    t5 = time.time()

    for i in range(estimator_count):
        e = Ridge(10, solver="lsqr")
        e.fit(X, y)

    t6 = time.time()

    for i in range(estimator_count):
        e = Ridge(10, solver="sag")
        e.fit(X, y)

    t7 = time.time()

    print()
    print(t2 - t1, t3 - t2, t4 - t3, t5 - t4, t6 - t5, t7 - t6)
    # neighbour = 500 0.5866434574127197 4.114291429519653 0.6661701202392578 0.4684324264526367 0.3340754508972168 9.437297821044922
    # neighbour = 1000 2.821377992630005 21.683172464370728 2.867004156112671 0.9687221050262451 0.9672167301177979 64.40824103355408


def test_alpha():
    neighbour_count = 1000
    X = np.random.random((neighbour_count, neighbour_count))
    y = np.random.random(neighbour_count)
    estimator_count = 100

    t1 = time.time()

    for i in range(estimator_count):
        e = Ridge(0)
        e.fit(X, y)

    t2 = time.time()

    for i in range(estimator_count):
        e = Ridge(0.1)
        e.fit(X, y)

    t3 = time.time()

    for i in range(estimator_count):
        e = Ridge(10)
        e.fit(X, y)

    t4 = time.time()

    print()
    print(t2 - t1, t3 - t2, t4 - t3)
    # neighbour = 1000 57.08750033378601 1.9294869899749756 1.918003797531128


if __name__ == "__main__":
    test_regression()
    # test_solver()
    # test_alpha()



from time import time

import numpy as np
from numba import jit, njit, prange
from sklearn.linear_model import Ridge


def loop_python(iteration_count=1000):
    X = np.random.random((100, 100))
    y = np.random.random((100, 1))

    for i in range(iteration_count):
        estimator = Ridge(1)
        estimator.fit(X, y)


@jit(forceobj=True, looplift=True)
def loop_jitting(iteration_count=1000):
    X = np.random.random((100, 100))
    y = np.random.random((100, 1))

    for i in range(iteration_count):
        estimator = Ridge(1)
        estimator.fit(X, y)


@njit()
def loop_numba(iteration_count=1000):
    X = np.random.random((100, 100))
    y = np.random.random((100, 1))

    for i in range(iteration_count):
        ridge_fit(X, y)


@njit(parallel=True)
def loop_paralle(iteration_count=1000):
    X = np.random.random((100, 100))
    y = np.random.random((100, 1))

    # Stuck when X, y is passed to ridge_fit. Everything is fine if X, y are generated inside ridge_fit.

    for i in prange(iteration_count):
        ridge_fit(X, y)


# @njit(parallel=True)
def loop_paralle_lstsq(iteration_count=1000):
    X = np.random.random((100, 100))
    y = np.random.random((100, 1))

    for i in prange(iteration_count):
        alpha = 1.0

        # Center the data to make the intercept term zero
        X_offset = mean(X, axis=0)
        y_offset = mean(y, axis=0)
        X_center = X - X_offset
        y_center = y - y_offset

        dimension = X_center.shape[1]
        A = np.identity(dimension)

        X_aug = np.vstack((X_center, np.sqrt(alpha) * A))
        y_aug = np.vstack((y_center, np.zeros((A.shape[0], 1), dtype=y.dtype)))

        coef = np.linalg.lstsq(X_aug, y_aug)[0]
        intercept = y_offset - np.dot(X_offset, coef)


@njit(parallel=True)
def loop_parallel_chol(iteration_count=1000):
    X = np.random.random((100, 100))
    y = np.random.random((100, 1))

    for i in prange(iteration_count):
        ridge_cholesky(X, y)

        # alpha = 1.0
        #
        # # Center the data to make the intercept term zero
        # X_offset = mean(X, axis=0)
        # y_offset = mean(y, axis=0)
        # X_center = X - X_offset
        # y_center = y - y_offset
        #
        # A = np.dot(X_center.T, X_center)
        # Xy = np.dot(X_center.T, y_center)
        #
        # A = A + alpha * np.eye(X.shape[1])
        #
        # coef = np.linalg.solve(A, Xy)
        # intercept = y_offset - np.dot(X_offset, coef)

@njit(parallel=True)
def loop_parallel_inner(iteration_count=1000):
    X = np.random.random((100, 100))
    y = np.random.random((100, 1))

    # coef_list = []
    # intercept_list = []

    for i in prange(iteration_count):
        alpha = 1.0
        
        # Center the data to make the intercept term zero
        # X_offset = mean(X, axis=0)
        # y_offset = mean(y, axis=0)
        # X_center = X - X_offset
        # y_center = y - y_offset

        X_center = X
        y_center = y
        
        dimension = X_center.shape[1]
        A = np.identity(dimension)
        A_biased = alpha * A

        temp = X_center.T.dot(X_center) + A_biased

        np.linalg.inv(
            temp
        )

        # coef = np.linalg.inv(
        #     X_center.T.dot(X_center) + A_biased
        # ).dot(X_center.T).dot(y_center)
        # intercept = y_offset - np.dot(X_offset, coef)
        
        # coef_list.append(coef)
        # intercept_list.append(intercept)


@njit()
def mean(x, axis, weight):
    weight = weight.reshape((-1, 1))
    x = x * weight
    weight = weight.reshape((1, -1))

    return np.sum(x, axis) / weight.sum()


@njit()
def ridge_fit(X, y):
    alpha = 1.0

    # Center the data to make the intercept term zero
    X_offset = mean(X, axis=0)
    y_offset = mean(y, axis=0)
    X_center = X - X_offset
    y_center = y - y_offset

    dimension = X_center.shape[1]
    A = np.identity(dimension)
    A_biased = alpha * A

    coef = np.linalg.inv(X_center.T.dot(X_center) + A_biased).dot(X_center.T).dot(y_center)
    intercept = y_offset - np.dot(X_offset, coef)

    return coef, intercept


@njit()
def rigde_lstsq(X, y):
    alpha = 1.0

    # Center the data to make the intercept term zero
    X_offset = mean(X, axis=0)
    y_offset = mean(y, axis=0)
    X_center = X - X_offset
    y_center = y - y_offset

    dimension = X_center.shape[1]
    A = np.identity(dimension)

    X_aug = np.vstack((X_center, np.sqrt(alpha) * A))
    y_aug = np.vstack((y_center, np.zeros((A.shape[0], 1), dtype=y.dtype)))

    coef = np.linalg.lstsq(X_aug, y_aug)[0]
    intercept = y_offset - np.dot(X_offset, coef)

    return coef, intercept

@njit()
def ridge_cholesky(X, y, weights=None):
    alpha = 1.0

    # Center the data to make the intercept term zero

    # TODO: Weight
    X_offset = mean(X, axis=0, weight=weights)
    y_offset = mean(y, axis=0, weight=weights)

    X_center = X - X_offset
    y_center = y - y_offset

    if weights is not None:
        weights_sqrt = np.sqrt(weights)
        for index, weight in enumerate(weights_sqrt):
            X_center[index] *= weight
            y_center[index] *= weight

    A = np.dot(X_center.T, X_center)
    Xy = np.dot(X_center.T, y_center)

    A = A + alpha * np.eye(X.shape[1])

    coef = np.linalg.solve(A, Xy)
    intercept = y_offset - np.dot(X_offset, coef)

    return coef, intercept


def test_ridge_work():
    X = np.random.random((1000, 100))
    y = np.random.random((1000, 1))
    weight = np.random.random((1000, ))

    # ridge_fit(X, y)
    t1 = time()
    # coef, intercept = ridge_fit(X, y)
    t2 = time()
    # print(coef, intercept)

    t3 = time()
    estimator = Ridge(1.0).fit(X, y, sample_weight=weight)
    t4 = time()
    print(estimator.coef_, estimator.intercept_)

    # rigde_lstsq(X, y)
    t5 = time()
    # coef, intercept = rigde_lstsq(X, y)
    t6 = time()
    # print(coef, intercept)

    coef, intercept = ridge_cholesky(X, y, weight)
    t7 = time()
    coef, intercept = ridge_cholesky(X, y, weight)
    t8 = time()
    print(coef, intercept)


    print(t2 - t1)
    print(t4 - t3)
    print(t6 - t5)
    print(t8 - t7)


def test_loop():

    t1 = time()
    loop_python(100000)
    t2 = time()
    print(t2 - t1)

    # loop_jitting()
    t1 = time()
    # loop_jitting()
    t2 = time()
    print(t2 - t1)

    # loop_numba()
    t1 = time()
    # loop_numba()
    t2 = time()
    print(t2 - t1)

    # loop_paralle()
    t1 = time()
    # loop_paralle()
    t2 = time()
    print(t2 - t1)

    # loop_parallel_inner()
    t1 = time()
    # loop_parallel_inner()
    t2 = time()
    print(t2 - t1)

    # loop_paralle_lstsq(1)
    t1 = time()
    # loop_paralle_lstsq(10000)
    t2 = time()
    print(t2 - t1)

    loop_parallel_chol(1)
    t1 = time()
    loop_parallel_chol(100000)
    t2 = time()
    print(t2 - t1)

if __name__ == "__main__":
    test_ridge_work()

    # 41 for intel extension
    # 39 for original sklearn
    # test_loop()



from pathlib import Path
import matplotlib.pyplot as plt

# Create folder for saving plot
default_folder = Path('Plot')
default_folder.mkdir(exist_ok=True)

# Set default style for matplotlib

# Control style to make it consistent with the Plotly
# plt.style.use('seaborn')
plt.style.use('seaborn-v0_8')
# Font family
plt.rcParams["font.family"] = "Times New Roman"
plt.rcParams["font.size"] = 12
plt.rcParams["axes.labelsize"] = 12



from time import time as t

from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression, Ridge, RidgeCV
from sklearn.tree import DecisionTreeRegressor

from georegression.stacking_model import StackingWeightModel
from georegression.test.data import load_HP, load_ESI
from georegression.weight_model import WeightModel

X, y_true, xy_vector, time = load_ESI()


def test_large_data():
    local_estimator = DecisionTreeRegressor(splitter="random", max_depth=2)
    distance_measure = "euclidean"
    kernel_type = "bisquare"
    distance_ratio = None
    bandwidth = None
    neighbour_count = 0.01

    model = StackingWeightModel(
        local_estimator,
        distance_measure,
        kernel_type,
        distance_ratio,
        bandwidth,
        neighbour_count,
    )

    t_start = t()

    model.fit(X, y_true, [xy_vector, time])
    print(f"{model.llocv_score_}, {model.llocv_stacking_}")

    t_end = t()
    print(f"Time: {t_end - t_start}")

if __name__ == '__main__':
    test_large_data()


from time import time as t

from scipy.sparse import csr_array
from sklearn.tree import ExtraTreeRegressor

from georegression.test.data import load_HP
from georegression.weight_matrix import weight_matrix_from_points
from georegression.weight_model import WeightModel

X, y_true, xy_vector, time = load_HP()


def test_performance_n_jobs():
    weight_matrix = weight_matrix_from_points(
        [xy_vector, time],
        [xy_vector, time],
        "euclidean",
        "bisquare",
        neighbour_count=0.1,
    )

    t1 = t()
    estimator = WeightModel(
        ExtraTreeRegressor(max_depth=1, splitter="random"),
        n_jobs=-1
    )
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    t2 = t()
    print(t2 - t1)
    print(estimator.llocv_score_)

    weight_matrix = csr_array(weight_matrix)

    t1 = t()
    estimator = WeightModel(
        ExtraTreeRegressor(max_depth=1, splitter="random"),
        n_jobs=-1
    )
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    t2 = t()
    print(t2 - t1)
    print(estimator.llocv_score_)

    # 5.687727212905884
    # 0.7720929346825464
    # 1.841477394104004
    # 0.7732306074996614

def test_performance_n_patches():
    weight_matrix = weight_matrix_from_points(
        [xy_vector, time],
        [xy_vector, time],
        "euclidean",
        "bisquare",
        neighbour_count=0.1,
    )
    
    t1 = t()
    estimator = WeightModel(ExtraTreeRegressor(max_depth=1, splitter="random"), n_patches=6)
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    t2 = t()
    print(t2 - t1)
    print(estimator.llocv_score_)
    
    weight_matrix = csr_array(weight_matrix)
    
    t1 = t()
    estimator = WeightModel(ExtraTreeRegressor(max_depth=1, splitter="random"), n_patches=6)
    estimator.fit(X, y_true, [xy_vector, time], weight_matrix=weight_matrix)
    t2 = t()
    print(t2 - t1)
    print(estimator.llocv_score_)

    # 2.56437611579895
    # 0.7760204788183097
    # 0.9893820285797119
    # 0.7697055618854263

if __name__ == "__main__":
    test_performance_n_jobs()
    test_performance_n_patches()



import numpy as np
from georegression.simulation.simulation_for_fitting import generate_sample, f_square, coef_strong

X, y, points = generate_sample(500, f_square, coef_strong, random_seed=1, plot=True)
X_plus = np.concatenate([X, points], axis=1)

# --- Context ---

from sklearn.linear_model import LinearRegression
from georegression.weight_model import WeightModel

distance_measure = "euclidean"
kernel_type = "bisquare"

gwr_neighbour_count=0.2
model = WeightModel(
    LinearRegression(),
    distance_measure,
    kernel_type,
    neighbour_count=gwr_neighbour_count,
)
model.fit(X_plus, y, [points])

print('GWR R2 Score: ', model.llocv_score_)

# --- Alternative ---

from sklearn.metrics import r2_score
y_predict = model.local_predict_
score = r2_score(y, y_predict)
print(score)



from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor
from georegression.test.data import load_HP
from georegression.visualize.pd import select_partials
from georegression.weight_model import WeightModel


X, y, xy_vector, time = load_HP()


def test_partial_dependence():
    model = WeightModel(
        LinearRegression(),
        distance_measure='euclidean',
        kernel_type='bisquare',
        neighbour_count=0.5,

        cache_data=True, cache_estimator=True
    )

    model.fit(X[:100, :10], y[:100], [xy_vector[:100], time[:100]])
    model.partial_dependence()
    model.local_ICE()


if __name__ == '__main__':
    test_partial_dependence()



import math
from typing import Union

import dask.array as da
import numpy as np

from slab_utils.quick_logger import logger

KERNEL_TYPE_ENUM = [
    "linear",
    "uniform",
    "gaussian",
    "exponential",
    "boxcar",
    "bisquare",
    "tricube",
]


def kernel_function(
    distance: np.ndarray,
    bandwidth: Union[float, list[float], np.ndarray],
    kernel_type: str,
) -> Union[np.ndarray, da.Array]:
    """
    Using kernel function to calculate the weight

    Args:
        distance: The distances to be weighted by kernel function. In vector or matrix form.

        bandwidth: parameter of the kernel function for specifying the decreasing level.
         For compact supported kernel, weight will be 0 if distance is larger than the bandwidth.

        kernel_type:

    Returns:
        np.ndarray: weight without normalization

    """

    # Reshape bandwidth to have proper broadcasting in division. It has been done in previous step.
    normalize_distance = distance / bandwidth
    if isinstance(normalize_distance, da.Array):
        # TODO: Why dask array do not support keyword argument for this method? It can be easily implemented with map_blocks
        # da.map_blocks(np.nan_to_num, normalize_distance, copy=False, nan=0.0, posinf=np.inf)
        normalize_distance = da.nan_to_num(normalize_distance, False, 0.0, np.inf)
    else:
        np.nan_to_num(normalize_distance, copy=False, nan=0.0, posinf=np.inf)

    # Continuous kernel
    if kernel_type == "uniform":
        weight = np.ones_like(normalize_distance)
    elif kernel_type == "gaussian":
        weight = np.exp(-0.5 * normalize_distance**2)
    elif kernel_type == "exponential":
        weight = np.exp(-0.5 * np.abs(normalize_distance))

    # Compact supported kernel
    elif kernel_type == "linear":
        weight = 1 - normalize_distance
    elif kernel_type == "boxcar":
        weight = np.ones_like(normalize_distance)
    elif kernel_type == "bisquare":
        # Optimize for dask array
        if isinstance(normalize_distance, da.Array):
            normalize_distance[normalize_distance > 1] = 1
            weight = (1 - normalize_distance**2) ** 2

            # It will never be negative!!!
            # weight[weight < 0] = 0

            return weight

        weight = (1 - normalize_distance**2) ** 2
    elif kernel_type == "tricube":
        weight = (1 - np.abs(normalize_distance) ** 3) ** 3
    else:
        raise Exception("Unsupported kernel")

    # compact support
    if kernel_type in ["linear", "boxcar", "bisquare", "tricube"]:
        weight[distance > bandwidth] = 0

    return weight


def adaptive_bandwidth(distance: np.ndarray, neighbour_count: Union[int, float], distance_sorted: da.array) -> float:
    """
    Find the bandwidth to include the specified number of neighbour.

    Args:
        distance: The distances to calculate the adaptive bandwidth. In vector or matrix form.
        neighbour_count: Number of the neighbour to include by the bandwidth.
         Use float to specify the percentage of the neighbour to include.

    Returns:
        float: return the distance to the K nearest neighbour
    """

    if isinstance(distance, da.Array):
        # Support for dask array
        # Duplicated coordinate is not supported
        N = distance.shape[0]

        if distance_sorted is not None:
            # Must be the sorted distance matrix
            if isinstance(neighbour_count, float):
                neighbour_count = math.ceil(N * neighbour_count)

            if neighbour_count > N:
                raise Exception("Invalid neighbour count")

            bandwidth = distance_sorted[:, [neighbour_count - 1]]
            return bandwidth
        else:
            logger.warning('Sorted distance matrix is not provided, requiring longer time to perform quantile function')

            bandwidth = distance.map_blocks(
                np.quantile,
                neighbour_count,
                axis=1,
                keepdims=True,
                drop_axis=1,
                new_axis=1,
            )
            return bandwidth

    if neighbour_count <= 0:
        raise Exception("Invalid neighbour count")

    if isinstance(neighbour_count, float):
        # percentile call the partition function internally,
        bandwidth = np.quantile(distance, neighbour_count, axis=1, keepdims=True, method='median_unbiased')
    elif isinstance(neighbour_count, int):
        bandwidth = np.partition(distance, neighbour_count - 1)[:, [neighbour_count - 1]]
    else:
        raise Exception("Invalid neighbour count")

    return bandwidth


def adaptive_kernel(
        distance: np.ndarray,
        neighbour_count: Union[int, float],
        kernel_type: str,
        distance_sorted: da.array=None
) -> np.ndarray:
    """
    Deduce the bandwidth from the neighbour count and calculate weight using kernel function.

    Args:
        distance:
        neighbour_count:
        kernel_type:

    Returns:

    """

    bandwidth = adaptive_bandwidth(distance, neighbour_count, distance_sorted)
    return kernel_function(distance, bandwidth, kernel_type)



"""
References: https://github.com/SeldonIO/alibi/blob/master/alibi/explainers/ale.py
"""

from functools import partial
from typing import Callable, Tuple

import numpy as np


def get_quantiles(values: np.ndarray, num_quantiles: int = 11, interpolation='linear') -> np.ndarray:
    """
    Calculate quantiles of values in an array.

    Parameters
    ----------
    values
        Array of values.
    num_quantiles
        Number of quantiles to calculate.

    Returns
    -------
    Array of quantiles of the input values.

    """
    percentiles = np.linspace(0, 100, num=num_quantiles)
    quantiles = np.percentile(values, percentiles, axis=0, interpolation=interpolation)  # type: ignore[call-overload]
    return quantiles


def bisect_fun(fun: Callable, target: float, lo: int, hi: int) -> int:
    """
    Bisection algorithm for function evaluation with integer support.

    Assumes the function is non-decreasing on the interval `[lo, hi]`.
    Return an integer value v such that for all `x<v, fun(x)<target` and for all `x>=v, fun(x)>=target`.
    This is equivalent to the library function `bisect.bisect_left` but for functions defined on integers.

    Parameters
    ----------
    fun
        A function defined on integers in the range `[lo, hi]` and returning floats.
    target
        Target value to be searched for.
    lo
        Lower bound of the domain.
    hi
        Upper bound of the domain.

    Returns
    -------
    Integer index.

    """
    while lo < hi:
        mid = (lo + hi) // 2
        if fun(mid) < target:
            lo = mid + 1
        else:
            hi = mid
    return lo


def minimum_satisfied(values: np.ndarray, min_bin_points: int, n: int) -> int:
    """
    Calculates whether the partition into bins induced by `n` quantiles
    has the minimum number of points in each resulting bin.

    Parameters
    ----------
    values
        Array of feature values.
    min_bin_points
        Minimum number of points each discretized interval needs to contain.
    n
        Number of quantiles.

    Returns
    -------
    Integer encoded boolean with 1 - each bin has at least `min_bin_points` and 0 otherwise.

    """
    q = np.unique(get_quantiles(values, num_quantiles=n))
    indices = np.searchsorted(q, values, side='left')
    indices[indices == 0] = 1
    interval_n = np.bincount(indices)
    return int(np.all(interval_n[1:] > min_bin_points))


def adaptive_grid(values: np.ndarray, min_bin_points: int = 1) -> Tuple[np.ndarray, int]:
    """
    Find the optimal number of quantiles for the range of values so that each resulting bin
    contains at least `min_bin_points`. Uses bisection.

    Parameters
    ----------
    values
        Array of feature values.
    min_bin_points
        Minimum number of points each discretized interval should contain to ensure more precise
        ALE estimation.

    Returns
    -------
    q
        Unique quantiles.
    num_quantiles
        Number of non-unique quantiles the feature array was subdivided into.

    Notes
    -----
    This is a heuristic procedure since the bisection algorithm is applied
    to a function which is not monotonic. This will not necessarily find the
    maximum number of bins the interval can be subdivided into to satisfy
    the minimum number of points in each resulting bin.
    """

    # function to bisect
    def minimum_not_satisfied(values: np.ndarray, min_bin_points: int, n: int) -> int:
        """
        Logical not of `minimum_satisfied`, see function for parameter information.
        """
        return 1 - minimum_satisfied(values, min_bin_points, n)

    fun = partial(minimum_not_satisfied, values, min_bin_points)

    # bisect
    num_quantiles = bisect_fun(fun=fun, target=0.5, lo=0, hi=len(values)) - 1
    q = np.unique(get_quantiles(values, num_quantiles=num_quantiles))

    return q, num_quantiles



import pandas as pd
import plotly.graph_objects as go
import numpy as np
from matplotlib import cm
import plotly.express as px


def value_to_rgb_str(value, colormap_name='viridis'):
    colormap = cm.get_cmap(colormap_name)
    rgba_tuple = colormap(value)
    return f'rgb({int(rgba_tuple[0] * 255)}, {int(rgba_tuple[1] * 255)}, {int(rgba_tuple[2] * 255)})'


def value_list_to_rgb_str_list(value_list):
    return [
        value_to_rgb_str(value)
        for value in value_list
    ]


colors = px.colors.sequential.RdBu


def test_mesh_3d():
    fig = go.Figure(data=[
        # Point
        go.Scatter3d(x=[0, 1, 1, 0], y=[0, 0, 1, 1], z=[0, 0, 0, 0], hovertext=[1, 2, 3, 4],
                     marker={
                         # 'colorscale': px.colors.qualitative.Plotly,
                         # 'colorscale': value_list_to_rgb_str_list([0, 0.25, 0.25, 0.88]),
                         'colorscale': colors,

                         'colorbar': {
                             'x': 1,
                             'tickformat': '.4s',
                             # 'tickvals': value_list_to_rgb_str_list([0, 0.25, 0.25, 0.88])
                             # 'tickvals': [0, 1]
                         },
                         # 'color': value_list_to_rgb_str_list([0, 0.25, 0.25, 0.88])
                         'color': [0, 0.25, 0.25, 0.88]
                     }),
        # Line
        go.Scatter3d(x=[0, 1, 1, 0], y=[0, 0, 1, 1], z=[0, 1, 1, 0], hovertext=[1, 2, 3, 4],
                     mode='lines',
                     line={
                         # 'colorbar': {'x': 1.1},
                         # 'color': [1, 2, 3, 4]
                     }),
        # Bottom surface
        go.Mesh3d(x=[0, 1, 1, 0], y=[0, 0, 1, 1], z=[0, 0, 0, 0], color='rgb(0, 255, 0)', hoverinfo='skip',
                  hovertemplate=None, colorbar={'x': 1.1}),
        # Top surface
        go.Mesh3d(x=[0, 1, 1, 0], y=[0, 0, 1, 1], z=[1] * 4, color='rgb(255, 255, 0)', hoverinfo='skip',
                  hovertemplate=None)
    ])
    fig.update_layout(
        # Set axis ratio
        scene_aspectmode='manual',
        scene_aspectratio=dict(x=1, y=1, z=3),
        # Clear margin
        # automargin=True
        margin=dict(l=0, r=0, t=50, b=0, pad=0),
        hovermode='x',
        # hoverdistance=-1
        spikedistance=1
    )

    fig.update_layout(
        title={
            'text': f"Temporal Partial Dependency of Feature",
            'y': 0.95,  # new
            'x': 0.4,
            'xanchor': 'center',
            'yanchor': 'top'  # new
        },
        scene=dict(
            xaxis_title='X Axis Title',
            yaxis_title='Y Axis Title',
            zaxis_title='Z Axis Title',
        ),
        legend_title="Cluster Legend",
        font=dict(
            size=18,
            color="RebeccaPurple"
        ))

    # fig.update_xaxes(automargin=True)
    # fig.update_yaxes(automargin=True)

    fig2 = px.scatter_3d(x=[0, 1, 1, 0], y=[0, 0, 1, 1], z=[0, 0, 0, 0],
                         color=np.array([0, 0.25, 0.25, 0.88]).astype(str))
    fig.add_traces(fig2.data)

    fig.write_html(f'test_plot.html')


def test_discrete():
    fig = px.scatter_3d(x=[0, 1, 1, 0], y=[0, 0, 1, 1], z=[0, 0, 0, 0],
                        color=['T1', 'T2', 'T3', 'T4'])
    fig.write_html(f'test_plot2.html')


def test_lengend():
    import plotly.express as px
    df = px.data.gapminder()
    fig = px.line(df, y="lifeExp", x="year", color="continent", line_group="country",
                  line_shape="spline", render_mode="svg",
                  color_discrete_sequence=px.colors.qualitative.G10,
                  title="Built-in G10 color sequence")

    fig.show()


def test_surface_colorbar():
    fig = go.Figure([
        go.Mesh3d(
            x=[0, 1, 1, 0], y=[0, 0, 1, 1], z=[0, 0, 0, 0], color='rgb(0, 255, 0)', hoverinfo='skip',
            hovertemplate=None,
            name='Surface Group',
            legendgroup='surface group',
            showlegend=True,
            showscale=True,
            colorbar={'x': 1.1}
        ),
        go.Mesh3d(
            x=[0, 1, 1, 0], y=[0, 0, 1, 1], z=[1, 1, 1, 1], hoverinfo='skip',
            hovertemplate=None,
            name='surface2',
            legendgroup='surface group',
            showlegend=False,
            # colorbar={'x': 1.1}
        )
    ])

    fig.update_layout(
        scene=dict(
            xaxis=dict(
                backgroundcolor="rgb(200, 200, 230)",
                gridcolor="white",
            ),
        )
    )

    fig.write_html(f'test_plot.html')


def test_axes_ratio():
    import plotly.graph_objects as go

    fig = go.Figure()

    fig.add_trace(go.Scatter(
        x=[0, 1, 1, 0, 0, 1, 1, 2, 2, 3, 3, 2, 2, 3],
        y=[0, 0, 1, 1, 3, 3, 2, 2, 3, 3, 1, 1, 0, 0]
    ))

    fig.update_layout(
        width=800,
        height=500,
        title="fixed-ratio axes"
    )
    fig.update_xaxes(
        range=(-0.5, 3.5),
        constrain='domain'
    )
    fig.update_yaxes(
        scaleanchor="x",
        scaleratio=1,
    )

    fig.show()

if __name__ == '__main__':
    pass



import numpy as np
from scipy.stats import multivariate_normal

def gaussian_coefficient(mean, cov, amplitude=1):
    """
    Generate a gaussian coefficient for a given mean and variance.
    """

    if isinstance(cov, list):
        cov = np.array(cov)

    normalize_factor = multivariate_normal.pdf(mean, mean, cov)

    def coefficient(point):
        return multivariate_normal.pdf(point, mean, cov) * amplitude / normalize_factor

    return coefficient


def radial_coefficient(origin, amplitude):
    """
    Generate a radial coefficient for a given origin.
    """

    def coefficient(point):
        return np.linalg.norm(point - origin, axis=-1) * amplitude

    return coefficient


def directional_coefficient(direction, amplitude=1):
    """
    Generate a directional coefficient for a given direction.
    """

    # Normalize the direction vector
    direction = direction / np.linalg.norm(direction)

    def coefficient(point):
        return np.dot(point, direction) / np.linalg.norm(point, axis=-1) * amplitude

    return coefficient


def sine_coefficient(frequency, direction, amplitude):
    """
    Generate a sine coefficient for a given frequency.
    """

    def coefficient(point):
        return np.sin(np.dot(point, direction) * frequency) * amplitude

    return coefficient


def coefficient_wrapper(operator, *coefficients):
    """
    Wrap two coefficients with an operator.
    """
    if len(coefficients) == 1:
        def coefficient_func(point):
            return operator(coefficients[0](point))

        return coefficient_func

    elif coefficients is not None:
        def coefficient_func(point):
            return operator(np.column_stack([c(point) for c in coefficients]), axis=-1)

        return coefficient_func

    return None


def polynomial_function(coefficient, degree):
    def function(x, point):
        return coefficient(point) * x ** degree

    return function


def square_function(coefficient):
    return polynomial_function(coefficient, 2)


def linear_function(coefficient):
    return polynomial_function(coefficient, 1)


def interaction_function(coefficient):
    def function(x1, x2, point):
        return coefficient(point) * x1 * x2

    return function


def sigmoid_function(coefficient):
    def function(x, point):
        return coefficient(point) * np.tanh(x)

    return function


def relu_function(coefficient):
    def function(x, point):
        return coefficient(point) * np.maximum(0, x)

    return function


def exponential_function(coefficient):
    def function(x, point):
        return coefficient(point) * np.exp(x)

    return function

def coefficient_function(coefficient):
    def function(point):
        return coefficient(point)
    
    return function


def sample_points(n, dim=2, bounds=(-1, 1), numeraical_type="continuous"):
    """
    Sample n points in dim dimensions from a uniform distribution with given bounds.
    The bound of each dimension can be sampled from continuous range or discrete classes.
    """

    points = np.zeros((n, dim))

    if not isinstance(bounds, list):
        bounds = [bounds] * dim

    if not isinstance(numeraical_type, list):
        numeraical_type = [numeraical_type] * dim

    for i in range(dim):
        if numeraical_type[i] == "continuous":
            points[:, i] = np.random.uniform(bounds[i][0], bounds[i][1], n)
        elif numeraical_type[i] == "discrete":
            points[:, i] = np.random.choice(bounds[i], n)

    return points


def sample_x(n, type='uniform', bounds=(-1, 1), mean=0, variance=1, scale=1, points=None):
    """
    Sample n x values from a specified distribution.
    If points are given, the parameter of distribution can be determined by coefficient functions,
    which introduce the spatial dependency.
    """
    if callable(bounds[0]):
        bounds[0] = bounds[0](points)

    if callable(bounds[1]):
        bounds[1] = bounds[1](points)

    if callable(mean):
        mean = mean(points)
    
    if callable(variance):
        variance = variance(points)

    if callable(scale):
        scale = scale(points)


    if type == 'uniform':
        base = np.random.uniform(0, 1, n)
        return base * (bounds[1] - bounds[0]) + bounds[0] + mean
    elif type == 'normal':
        base = np.random.normal(0, 1, n)
        return base * variance + mean
    elif type == 'exponential':
        base = np.random.exponential(1, n)
        return base * scale



from georegression.test.data import load_HP
from georegression.test.visualize import get_toy_model
from georegression.visualize.pd import partials_plot_3d, features_partial_cluster, partial_plot_2d, \
    partial_compound_plot, choose_cluster_typical

model = get_toy_model()
features_embedding, features_cluster_label, _ = features_partial_cluster(model.feature_partial_)

X, y, xy_vector, time = load_HP()


def test_compound_plot():
    partial_figs, embedding_figs, cluster_figs, compass_figs = partial_compound_plot(
        xy_vector[:100], time[:100], model.feature_partial_,
        features_embedding, features_cluster_label,
    )


def test_pd_2d_plot():
    # cluster_typical = choose_cluster_typical(cluster_embedding, cluster_label)
    # partial_plot_2d(
    #     model.feature_partial_, cluster_label, cluster_typical,
    #     alpha_range=[0.1, 1], width_range=[0.5, 3], scale_power=1.5
    # )

    cluster_typical = [
        choose_cluster_typical(embedding, cluster)
        for embedding, cluster in zip(features_embedding, features_cluster_label)
    ]
    partial_plot_2d(
        model.feature_partial_, features_cluster_label, cluster_typical,
        alpha_range=[0.3, 1], width_range=[0.5, 3], scale_power=1.5
    )


def test_pd_3d_plot():
    partials_plot_3d(
        model.feature_partial_, model.coordinate_vector_list[1], cluster_labels=features_cluster_label,
        # quantile=[0, 0.2, 0.8, 1],
    )


if __name__ == '__main__':
    test_pd_2d_plot()
    pass



"""
References: https://github.com/SeldonIO/alibi/blob/master/alibi/explainers/ale.py
"""

import numpy as np
import pandas as pd

from georegression.ale_utils import adaptive_grid


def weighted_ale(X, feature, predictor, weights=None, normalize=False, min_bin_points=5):
    fvals, _ = adaptive_grid(X[:, feature], min_bin_points)

    # find which interval each observation falls into
    indices = np.searchsorted(fvals, X[:, feature], side="left")
    indices[indices == 0] = 1  # put the smallest data point in the first interval
    interval_n = np.bincount(indices)  # number of points in each interval

    # predictions for the upper and lower ranges of intervals
    z_low = X.copy()
    z_high = X.copy()
    z_low[:, feature] = fvals[indices - 1]
    z_high[:, feature] = fvals[indices]
    p_low = predictor(z_low)
    p_high = predictor(z_high)

    # finite differences
    p_deltas = p_high - p_low

    # base value, which is the average prediction for the lowest interval
    base_value = np.average(p_low[indices == 1], weights=weights[indices == 1])

    # make a dataframe for averaging over intervals
    concat = np.column_stack((p_deltas, indices, weights))
    df = pd.DataFrame(concat)

    # weighted average for each interval
    avg_p_deltas = df.groupby(1).apply(lambda x: np.average(x[0], weights=x[2])).values

    # accumulate over intervals
    accum_p_deltas = np.cumsum(avg_p_deltas, axis=0)

    # pre-pend 0 for the left-most point
    zeros = np.zeros((1, 1))
    accum_p_deltas = np.insert(accum_p_deltas, 0, zeros, axis=0)

    # center
    if normalize:
        # mean effect, R's `ALEPlot` and `iml` version (approximation per interval)
        # Eq.16 from original paper "Visualizing the effects of predictor variables in black box supervised learning models"
        ale0 = (
                0.5 * (accum_p_deltas[:-1] + accum_p_deltas[1:]) * interval_n[1:]
        ).sum(axis=0)
        ale0 = ale0 / interval_n.sum()

        ale = accum_p_deltas - ale0
    else:
        ale = accum_p_deltas + base_value

    return fvals, ale



from pathlib import Path

import numpy as np
from plotly import express as px, graph_objects as go
from georegression.visualize.utils import vector_to_color

from georegression.visualize import default_folder


def scatter_3d(
        geo_vector, temporal_vector, value,
        figure_title, value_name, filename=None,
        is_cluster=False,
        folder=default_folder
):
    # Shape(N, )
    x = geo_vector[:, 0]
    y = geo_vector[:, 1]
    z = temporal_vector[:, 0]

    x_min = np.min(x)
    x_max = np.max(x)
    x_interval = x_max - x_min
    y_min = np.min(y)
    y_max = np.max(y)
    y_interval = y_max - y_min
    z_min = np.min(z)
    z_max = np.max(z)
    z_interval = z_max - z_min
    z_unique = np.unique(z)
    z_step = z_interval / len(z_unique)

    value = np.array(value)
    count = value.shape[0]
    value_min = np.min(value)
    value_max = np.max(value)
    value_interval = value_max - value_min

    # Index for each point
    custom_data = np.arange(count)

    # Multiple legend for cluster input.
    if is_cluster:
        # Quick way using express.
        # Maybe use the Graph Object to unify the style?
        # fig = px.scatter_3d(x=x, y=y, z=z, color=value.astype('str'))

        fig = go.Figure()
        color = vector_to_color(value)

        for cluster_value in np.unique(value):
            cluster_index = value == cluster_value
            fig.add_trace(
                go.Scatter3d(
                    x=x[cluster_index], y=y[cluster_index], z=z[cluster_index], mode='markers',
                    # Name of trace for legend display
                    name=f'Cluster {cluster_value}',
                    legendgroup=f'Cluster {cluster_value}',
                    marker={
                        'color': color[cluster_index],
                        'size': 5,
                    },
                    text=value[cluster_index],
                    customdata=custom_data,
                    hovertemplate=
                    f'<b>Time Slice</b> :' + ' %{z} <br />' +
                    f'<b>Index</b> :' + ' %{customdata} <br />' +
                    f'<b>{value_name}</b> :' + ' %{text} <br />' +
                    '<extra></extra>',
                )
            )
    # Continuous value case. Single legend/trace.
    else:
        tick_value = np.quantile(value, [0, 0.25, 0.5, 0.75, 1], interpolation='nearest')
        fig = go.Figure(data=[
            # Data Point
            go.Scatter3d(
                x=x, y=y, z=z, mode='markers',
                # Name of trace for legend display
                name=f'{value_name}',
                marker={
                    'size': 5,
                    'color': value,
                    'colorscale': 'Portland',
                    # Input dict of properties to construct the ColorBar Instance
                    'colorbar': {
                        'x': 0.8,
                        'title': f'{value_name} Color Bar<br>(Quartile tick)<br> <br>',
                        'tickvals': tick_value,
                        'tickformat': '.3~f',
                    },
                },
                text=value,
                customdata=custom_data,
                hovertemplate=
                '<b>Time Slice</b> :' + ' %{z} <br />' +
                f'<b>Index</b> :' + ' %{customdata} <br />' +
                f'<b>{value_name}</b> :' + ' %{text:.3~f} <br />' +
                '<extra></extra>'
            ),
        ])

    # TODO: Add Joint Line

    # Time Surface
    x_surface = (x_min, x_min, x_max, x_max)
    y_surface = (y_min, y_max, y_max, y_min)
    z_shift = z_step * 0.08
    surface_color = vector_to_color(np.unique(z))

    fig.add_traces([
        go.Mesh3d(
            x=x_surface, y=y_surface,
            # Shift the surface down a little to avoid overlay
            z=[z - z_shift] * 4, opacity=0.3,
            color=surface_color[z_index],
            hoverinfo='skip',
            name=f'Auxiliary Surface Group',
            legendgroup='Auxiliary Surface Group',
            showlegend=True if not z_index else False
        )
        for z_index, z in enumerate(z_unique)
    ])

    # Set figure, axis and other things.

    if x_interval < y_interval:
        x_aspect = 1
        y_aspect = y_interval / x_interval
        z_aspect = y_aspect * len(z_unique) * 0.6
    else:
        y_aspect = 1
        x_aspect = x_interval / y_interval
        z_aspect = x_aspect * len(z_unique) * 0.6

    fig.update_layout(
        # Clear margin
        margin=dict(l=0, r=0, t=50, b=0, pad=0),

        # Figure title
        title={
            'text': figure_title,
            'xanchor': 'center',
            'x': 0.45,
            'yanchor': 'top',
            'y': 0.99,
        },

        # Global font
        font=dict(size=12),

        # Legend
        legend_title="Point and Surface Legend",

        template="seaborn",
        font_family="Times New Roman"
    )

    fig.update_scenes(
        # Change Projection to Orthogonal
        camera_projection_type="orthographic",

        # Set axis ratio
        aspectmode='manual',
        aspectratio=dict(x=x_aspect, y=y_aspect, z=z_aspect),

        # Axis label
        xaxis=dict(
            title='X Position',
            ticktext=['Neg', 'Pos'],
            tickvals=[x_min, x_max],
            range=[x_min - x_interval * 0.12, x_max + x_interval * 0.12],
            # backgroundcolor="rgb(200, 200, 230)",
            # gridcolor="white",
            showbackground=True,
            # zerolinecolor="white",
            showspikes=False
        ),
        yaxis=dict(
            title='Y Position',
            ticktext=['Neg', 'Pos'],
            tickvals=[y_min, y_max],
            range=[y_min - y_interval * 0.12, y_max + y_interval * 0.12],
            # backgroundcolor="rgb(230, 200,230)",
            # gridcolor="white",
            showbackground=True,
            # zerolinecolor="white",
            showspikes=False
        ),
        zaxis=dict(
            title='Temporal Slice Index',
            tickvals=z_unique,
            range=[z_min - z_step * 0.5, z_max + z_step * 0.5],
            # backgroundcolor="rgb(230, 230,200)",
            # gridcolor="white",
            showbackground=True,
            # zerolinecolor="white",
            # showspikes=False
        ),
    )

    # Output to disk file
    if filename is None:
        filename = f'{figure_title}_{value_name}'

    fig.write_html(folder / f'{filename}.html')

    return fig



from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor
from georegression.test.data import load_HP
from georegression.weight_model import WeightModel
from georegression.visualize.ale import plot_ale

X, y, xy_vector, time = load_HP()


def test_ale():
    model = WeightModel(
        # LinearRegression(),
        RandomForestRegressor(n_estimators=10),
        distance_measure='euclidean',
        kernel_type='bisquare',
        neighbour_count=0.1,
        cache_data=True, cache_estimator=True, n_jobs=-1
    )

    model.fit(X[:, -5:], y, [xy_vector, time])
    ale_list = model.local_ALE(0)

    for ale in ale_list:
        fval, ale = ale
        plot_ale(fval, ale)

    print()



if __name__ == '__main__':
    test_ale()



import numpy as np
from matplotlib import colors, cm
from plotly import graph_objects as go


def color_to_str(color_vector):
    def inner_stringify(color):
        return np.array(
            f'rgba({int(color[0] * 255)},{int(color[1] * 255)},{int(color[2] * 255)},{int(color[3] * 255)})',
            dtype=object
        )

    return np.apply_along_axis(inner_stringify, -1, color_vector)


def vector_to_color(vector, stringify=True, colormap='viridis'):
    """
    Set style template will change the default colormap.

    Args:
        vector ():
        stringify ():
        colormap ():

    Returns:

    """
    vector_normed = (vector - np.min(vector, axis=-1, keepdims=True)) / (
            np.max(vector, axis=-1, keepdims=True) - np.min(vector, axis=-1, keepdims=True))
    color = cm.get_cmap(colormap)(vector_normed)
    if not stringify:
        return color
    return color_to_str(color)


def range_margin(vector=None, value_min=None, value_max=None, margin=0.05):
    if vector is None and (value_min is None or value_max is None):
        raise Exception('Invalid parameter')

    if vector is not None:
        value_min = np.min(vector)
        value_max = np.max(vector)

    interval = value_max - value_min

    return [value_min - interval * margin, value_max + interval * margin]


def aspect_ratio_scale(fig: go.Figure, scale: float = 0.2):
    x_aspect = fig.layout.scene.aspectratio.x
    y_aspect = fig.layout.scene.aspectratio.y
    z_aspect = fig.layout.scene.aspectratio.z

    fig.update_layout(
        scene_aspectratio={
            "x": x_aspect * scale,
            "y": y_aspect * scale,
            "z": z_aspect * scale,
        },
    )

    return fig


def camera_position(fig: go.Figure, rotation=-70, pitch=40):
    # From rotation to x y using sin and cos
    x = np.cos(rotation * np.pi / 180)
    y = np.sin(rotation * np.pi / 180)

    # From pitch to z using sin
    z = np.sin(pitch * np.pi / 180)
    camera = dict(
        eye=dict(x=x, y=y, z=z)
    )
    fig.update_layout(scene_camera=camera)

    return fig



from time import time

import numpy as np
from joblib import Parallel, delayed
from sklearn.linear_model import Ridge


def test_parallel_regression():
    # Generate stacking data
    neighbour_count = 100
    X = np.random.random((neighbour_count, neighbour_count))
    y = np.random.random(neighbour_count)
    estimator_count = 20000

    def fit_stacking():
        return Ridge().fit(X, y)

    t_start = time()
    job_list = [delayed(fit_stacking)() for i in range(estimator_count)]
    Parallel(n_jobs=-1)(job_list)
    t_end = time()

    print(t_end - t_start)
    # neighbour_count = 500 estimator_count = 100 1.5465185642242432
    # neighbour_count = 1000 estimator_count = 1000 7.498879909515381
    # neighbour_count = 20000 estimator_count = 500 21.9149112701416
    # neighbour_count = 100 estimator_count = 20000 1.5993006229400635


    t_start = time()
    for i in range(estimator_count):
        e = Ridge()
        e.fit(X, y)
    t_end = time()

    print(t_end - t_start)
    # neighbour_count = 1000 estimator_count = 1000 19.928532361984253
    # neighbour_count = 20000 estimator_count = 500 97.64331030845642
    # neighbour_count = 100 estimator_count = 20000 10.334396123886108




if __name__ == "__main__":
    test_parallel_regression()



import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestRegressor

X = np.random.normal(size=(100, 5))
y = np.random.normal(size=100)

df = pd.DataFrame(X, columns=["x1", "x2", "x3", "x4", "x5"])

estimator = RandomForestRegressor(n_estimators=100)
estimator.fit(X, y)


def PyALE():
    from PyALE import ale

    ale(df, estimator, ["x1"])


def alibiALE():
    from alibi.explainers import ALE
    from alibi.explainers import plot_ale

    def predict_fn(X):
        return estimator.predict(X)

    ale = ALE(predict_fn)
    exp = ale.explain(X)

    plot_ale(exp)


def ALEPython():
    from alepython import ale_plot

    # Plots ALE of feature 'cont' with Monte-Carlo replicas (default : 50).
    ale_plot(model, X_train, "cont", monte_carlo=True)


if __name__ == "__main__":
    alibiALE()



import json
import os
import time
from functools import partial

from sklearn.ensemble import RandomForestRegressor, ExtraTreesRegressor
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import LeaveOneOut
from sklearn.tree import DecisionTreeRegressor
from xgboost import XGBRegressor

from georegression.simulation.simulation import show_sample
from georegression.simulation.simulation_utils import *
from georegression.stacking_model import StackingWeightModel
from georegression.weight_model import WeightModel


# TODO: Explain why the improvement becomes significant when there are more points.


def fit_models(
    X,
    y,
    points,
    stacking_neighbour_count=0.03,
    stacking_neighbour_leave_out_rate=0.15,
    info=None,
):
    if info is None:
        info = {}

    X_plus = np.concatenate([X, points], axis=1)

    distance_measure = "euclidean"
    kernel_type = "bisquare"

    result = {}

    model = StackingWeightModel(
        DecisionTreeRegressor(splitter="random", max_depth=X.shape[1]),
        distance_measure,
        kernel_type,
        neighbour_count=stacking_neighbour_count,
        neighbour_leave_out_rate=stacking_neighbour_leave_out_rate,
    )

    repeats = 10
    stackings = []
    for i in range(repeats):
        t1 = time.time()
        model.fit(X_plus, y, [points])
        t2 = time.time()

        stackings.append(model.llocv_stacking_)

        print("Stacking:", model.llocv_score_, model.llocv_stacking_)
        print(t2 - t1)

        result[f"Stacking_Base_{i}"] = model.llocv_score_
        result[f"Stacking_{i}"] = model.llocv_stacking_
        result[f"Stacking_Time"] = t2 - t1

    result["Stacking_Mean"] = np.mean(stackings)
    result["Stacking_Std"] = np.std(stackings)
    result["Stacking_Max"] = np.max(stackings)
    result["Stacking_Min"] = np.min(stackings)

    model = StackingWeightModel(
        ExtraTreesRegressor(n_estimators=10, max_depth=X.shape[1]),
        distance_measure,
        kernel_type,
        neighbour_count=stacking_neighbour_count,
        neighbour_leave_out_rate=stacking_neighbour_leave_out_rate,
    )

    stackings_extra = []
    for i in range(repeats):
        t1 = time.time()
        model.fit(X_plus, y, [points])
        t2 = time.time()

        stackings_extra.append(model.llocv_stacking_)

        print("Stacking_Extra:", model.llocv_score_, model.llocv_stacking_)
        print(t2 - t1)

        result[f"Stacking_Extra_Base_{i}"] = model.llocv_score_
        result[f"Stacking_Extra_{i}"] = model.llocv_stacking_
        result[f"Stacking_Extra_Time"] = t2 - t1

    result["Stacking_Extra_Mean"] = np.mean(stackings_extra)
    result["Stacking_Extra_Std"] = np.std(stackings_extra)
    result["Stacking_Extra_Max"] = np.max(stackings_extra)
    result["Stacking_Extra_Min"] = np.min(stackings_extra)

    result = {**result, **info}
    with open("simulation_variation.jsonl", "a") as f:
        f.write(json.dumps(result) + "\n")

    return result


def coef_auto_gau_strong():
    coef_radial = radial_coefficient(np.array([0, 0]), 1 / np.sqrt(200))
    coef_dir = directional_coefficient(np.array([1, 1]))

    gau_coef_list = []
    for _ in range(1000):
        # Randomly generate the parameters for gaussian coefficient
        center = np.random.uniform(-10, 10, 2)
        amplitude = np.random.uniform(1, 2)
        sign = np.random.choice([-1, 1])
        amplitude *= sign
        sigma1 = np.random.uniform(0.2, 1)
        sigma2 = np.random.uniform(0.2, 1)
        cov = np.random.uniform(-np.sqrt(sigma1 * sigma2), np.sqrt(sigma1 * sigma2))
        sigma = np.array([[sigma1, cov], [cov, sigma2]])

        coef_gau = gaussian_coefficient(center, sigma, amplitude=amplitude)
        gau_coef_list.append(coef_gau)

    coef_gau = coefficient_wrapper(np.sum, *gau_coef_list)
    coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_dir, coef_gau)

    return coef_sum


def f_square(X, C, points):
    return polynomial_function(C[0], 2)(X[:, 0], points) + 0


def generate_sample(count, f, coef_func, random_seed=1, plot=False):
    np.random.seed(random_seed)
    points = sample_points(count, bounds=(-10, 10))
    x1 = sample_x(count, bounds=(-10, 10))
    coefficients = [coef_func()]

    X = np.stack((x1,), axis=-1)
    y = f(X, coefficients, points)

    if plot:
        folder = f"Plot/{coef_func.__name__}_{f.__name__}_{count}"
        os.makedirs(folder, exist_ok=True)
        show_sample(X, y, points, coefficients, folder)

    return X, y, points


def square_gau_strong_100():
    X, y, points = generate_sample(
        100, f_square, coef_auto_gau_strong, random_seed=1, plot=True
    )

    return fit_models(
        X,
        y,
        points,
        stacking_neighbour_count=0.45,
        stacking_neighbour_leave_out_rate=0.2,
        info={"f": "f_square", "coef": "coef_gau_strong", "count": 100},
    )


def square_gau_strong_500():
    X, y, points = generate_sample(
        500, f_square, coef_auto_gau_strong, random_seed=1, plot=True
    )

    return fit_models(
        X,
        y,
        points,
        stacking_neighbour_count=0.08,
        stacking_neighbour_leave_out_rate=0.1,
        info={"f": "f_square", "coef": "coef_gau_strong", "count": 500},
    )


def square_gau_strong_1000():
    X, y, points = generate_sample(
        1000, f_square, coef_auto_gau_strong, random_seed=1, plot=True
    )

    return fit_models(
        X,
        y,
        points,
        stacking_neighbour_count=0.02,
        stacking_neighbour_leave_out_rate=0.05,
        info={"f": "f_square", "coef": "coef_gau_strong", "count": 1000},
    )


def square_gau_strong_5000():
    X, y, points = generate_sample(
        5000, f_square, coef_auto_gau_strong, random_seed=1, plot=True
    )

    return fit_models(
        X,
        y,
        points,
        stacking_neighbour_count=0.008,
        stacking_neighbour_leave_out_rate=0.2,
        info={"f": "f_square", "coef": "coef_gau_strong", "count": 5000},
    )


def square_gau_strong_10000():
    X, y, points = generate_sample(
        10000, f_square, coef_auto_gau_strong, random_seed=1, plot=True
    )

    return fit_models(
        X,
        y,
        points,
        stacking_neighbour_count=0.008,
        stacking_neighbour_leave_out_rate=0.2,
        info={"f": "f_square", "coef": "coef_gau_strong", "count": 10000},
    )





if __name__ == "__main__":
    # square_gau_strong_100()
    # square_gau_strong_500()
    square_gau_strong_1000()
    # square_gau_strong_5000()
    # square_gau_strong_10000()

    pass



import numpy as np
from georegression.simulation.simulation_for_fitting import generate_sample, f_square, coef_strong

X, y, points = generate_sample(500, f_square, coef_strong, random_seed=1, plot=True)
X_plus = np.concatenate([X, points], axis=1)

# --- Context ---

from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import ExtraTreesRegressor
from georegression.stacking_model import StackingWeightModel

distance_measure = "euclidean"
kernel_type = "bisquare"

stacking_neighbour_count=0.3
stacking_neighbour_leave_out_rate=0.1
model = StackingWeightModel(
    DecisionTreeRegressor(splitter="random", max_depth=X.shape[1]),
    # Or use the ExtraTreesRegressor for better predicting performance.
    # ExtraTreesRegressor(n_estimators=10, max_depth=X.shape[1]), 
    distance_measure,
    kernel_type,
    neighbour_count=stacking_neighbour_count,
    neighbour_leave_out_rate=stacking_neighbour_leave_out_rate,
)
model.fit(X_plus, y, [points])
print('STST R2 Score: ', model.llocv_stacking_)

# --- Alternative ---

from sklearn.metrics import r2_score
y_predict = model.stacking_predict_
score = r2_score(y, y_predict)
print(score)



import matplotlib.pyplot as plt
import numpy as np
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeRegressor
from georegression.local_ale import weighted_ale

from georegression.stacking_model import StackingWeightModel
from georegression.simulation.simulation import generate_sample
from georegression.visualize.ale import plot_ale
from georegression.weight_model import WeightModel



def draw_graph():
    X, y, points, f, coef = generate_sample(count=5000, random_seed=1)
    X_plus = np.concatenate([X, points], axis=1)
    distance_measure = "euclidean"
    kernel_type = "bisquare"
    neighbour_count = 0.05

    # local_estimator = DecisionTreeRegressor(splitter="random", max_depth=1)
    # local_estimator = DecisionTreeRegressor(splitter="random", max_depth=2)
    # model = StackingWeightModel(
    #     local_estimator,
    #     distance_measure,
    #     kernel_type,
    #     neighbour_count=neighbour_count,
    #     neighbour_leave_out_rate=0.25,
    #     cache_data=True,
    #     cache_estimator=True,
    # )
    # model.fit(X, y, [points])
    # print('Stacking:', model.llocv_score_, model.llocv_stacking_)

    model = WeightModel(
        RandomForestRegressor(n_estimators=50),
        distance_measure,
        kernel_type,
        neighbour_count=neighbour_count,
        cache_data=True,
        cache_estimator=True,
    )
    model.fit(X, y, [points])
    print('GRF:', model.llocv_score_)

    feature_index = 0

    # ale_list = model.local_ALE(feature_index)
    for local_index in range(model.N):
        # fval, ale = ale_list[local_index]

        estimator = model.local_estimator_list[local_index]
        neighbour_mask = model.neighbour_matrix_[local_index]
        neighbour_weight = model.weight_matrix_[local_index][neighbour_mask]
        X_local = model.X[neighbour_mask]
        ale_result = weighted_ale(
            X_local, feature_index, estimator.predict, neighbour_weight)

        fval, ale = ale_result

        x_neighbour = X[model.neighbour_matrix_[local_index], feature_index]
        y_neighbour = y[model.neighbour_matrix_[local_index]]
        weight_neighbour = model.weight_matrix_[
            local_index, model.neighbour_matrix_[local_index]]
        
        # show_function_at_point(f, coef, points[local_index], ax=ax)
        # Get the true marginal effect for function f = x1 * x2.
        x_gird = np.linspace(np.min(x_neighbour), np.max(x_neighbour), 1000)
        x1 = X[local_index, 1]
        x1 = np.tile(x1, 1000)

        # X_grid = np.stack([x_gird, x1], axis=-1)
        # y_grid = f(X_grid, coef, points[local_index])

        from georegression.simulation.simulation import x2_coef
        beta = coef[0](points[local_index])
        x2_average = x2_coef(points[local_index])
        y_grid = beta * 0.5 * ((x2_average + 2) ** 2 -
                               (x2_average - 2) ** 2) * (1 / 4) * x_gird
        
        x1_base = np.empty(500)
        x1_base[:] = np.min(x_neighbour)
        x2_base = np.random.uniform(x2_average - 2, x2_average + 2, 500)
        base_value_real = estimator.predict(np.stack([x1_base, x2_base], axis=-1)).mean()

        diff = ale[0] - base_value_real
        ale = ale - diff

        fig = plot_ale(fval, ale, x_neighbour)
        ax = fig.get_axes()[0]
        scatter = ax.scatter(x_neighbour, y_neighbour, c=weight_neighbour)
        ax.scatter(X[local_index, feature_index], y[local_index], c='red', label='Local point')
        fig.colorbar(scatter, ax=ax, label='Weight')

        ax.plot(x_gird, y_grid, label="Function value")

        ale_result = weighted_ale(X_local, feature_index, estimator.predict, np.ones(X_local.shape[0]))
        fval, ale = ale_result
        diff = ale[0] - base_value_real
        ale = ale - diff
        # ax.plot(fval, ale, label="Neighbour ALE")


        # Add non-weighted ALE plot
        # ale_result = weighted_ale(X_local, feature_index, estimator.predict, np.ones(X_local.shape[0]))
        # Select the X that is in the value range of x_neighbour
        x_global_ale = X[(X[:, feature_index] >= np.min(x_neighbour)) & (X[:, feature_index] <= np.max(x_neighbour))]
        ale_result = weighted_ale(
            x_global_ale, feature_index, estimator.predict, np.ones(x_global_ale.shape[0]))
        fval, ale = ale_result

        x1_base = np.empty(500)
        x1_base[:] = np.min(x_neighbour)
        x2_base = np.random.choice(X[:, 1], 500)
        base_value_real = estimator.predict(np.stack([x1_base, x2_base], axis=-1)).mean()

        diff = ale[0] - base_value_real
        ale = ale - diff

        ax.plot(fval, ale, label="Non-weighted ALE")        

        # Add legend
        handles, labels = ax.get_legend_handles_labels()
        handles.append(scatter)
        labels.append('Weight')
        ax.legend(handles, labels)

        plt.show(block=True)


    importance_global = model.importance_score_global()
    print(importance_global)

    importance_local = model.importance_score_local()
    print(importance_local)

    # Normalize the local importance to [0, 1]
    importance_local = (importance_local - importance_local.min()) / \
        (importance_local.max() - importance_local.min())

    # Plot the local importance
    fig, ax = plt.subplots()
    scatter = ax.scatter(points[:, 0], points[:, 1],
                         c=importance_local[:, 0], cmap='viridis')
    fig.colorbar(scatter)
    plt.show()

    # Show the residual across the space.
    residual = model.stacking_predict_ - model.y_sample_
    residual = np.abs(residual)
    fig, ax = plt.subplots()
    # Lower residual values has lower transparency
    scatter = ax.scatter(points[:, 0], points[:, 1],
                         c=residual, alpha=residual / residual.max())
    fig.colorbar(scatter)
    fig.savefig('residual.png')

    fval, ale = model.global_ALE(feature_index)
    fig = plot_ale(fval, ale, X[:, feature_index])
    fig.savefig('ale_global.png')


if __name__ == "__main__":
    draw_graph()



import numpy as np
from georegression.simulation.simulation_for_fitting import generate_sample, f_square, coef_strong

X, y, points = generate_sample(500, f_square, coef_strong, random_seed=1, plot=True)
X_plus = np.concatenate([X, points], axis=1)

from sklearn.ensemble import RandomForestRegressor
from georegression.weight_model import WeightModel

# --- Context ---

times = np.random.randint(0, 10, size=(X.shape[0], 1))
X_plus = np.concatenate([X, points, times], axis=1)

distance_measure = ["euclidean", 'euclidean']
kernel_type = ["bisquare", 'bisquare']

grf_neighbour_count = 0.3

grf_n_estimators=50
model = WeightModel(
    RandomForestRegressor(n_estimators=grf_n_estimators),
    distance_measure,
    kernel_type,
    neighbour_count=grf_neighbour_count,
)
model.fit(X_plus, y, [points, times])


import numpy as np

from georegression.neighbour_utils import sample_neighbour


def test_sample_neighbour():
    # Generate ramdom weight matrix
    weight_matrix = np.random.rand(100, 100)

    # Set 0 randomly
    weight_matrix[weight_matrix < 0.99] = 0

    # Sample neighbour
    neighbour_matrix_sampled = sample_neighbour(weight_matrix, sample_rate=0.5)

if __name__ == '__main__':
    test_sample_neighbour()



import time
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
from sklearn.ensemble import RandomForestRegressor, ExtraTreesRegressor
from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeRegressor
from georegression.local_ale import weighted_ale

from georegression.stacking_model import StackingWeightModel
from georegression.simulation.simulation import show_sample
from georegression.visualize.ale import plot_ale
from georegression.weight_model import WeightModel
from georegression.simulation.simulation_utils import *


def coef_manual_gau():
    coef_radial = radial_coefficient(np.array([0, 0]), 1 / np.sqrt(200))
    coef_dir = directional_coefficient(np.array([1, 1]))

    coef_gau_1 = gaussian_coefficient(np.array([-5, 5]), [[3, 4], [4, 8]], amplitude=-1)
    coef_gau_2 = gaussian_coefficient(np.array([-2, -5]), 5, amplitude=2)
    coef_gau_3 = gaussian_coefficient(np.array([8, 3]), 10, amplitude=-1.5)
    coef_gau_4 = gaussian_coefficient(
        np.array([2, 8]), [[3, 0], [0, 15]], amplitude=0.8
    )
    coef_gau_5 = gaussian_coefficient(np.array([5, -10]), 1, amplitude=1)
    coef_gau_6 = gaussian_coefficient(np.array([-10, -10]), 15, amplitude=1.5)
    coef_gau_6 = gaussian_coefficient(np.array([-11, 0]), 5, amplitude=2)
    coef_gau_6 = gaussian_coefficient(np.array([-11, 0]), 5, amplitude=2)
    coef_gau = coefficient_wrapper(
        np.sum, coef_gau_1, coef_gau_2, coef_gau_3, coef_gau_4, coef_gau_5, coef_gau_6
    )

    # coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_dir, coef_gau)
    coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_gau)

    return coef_sum


def coef_auto_gau_weak():
    coef_radial = radial_coefficient(np.array([0, 0]), 1 / np.sqrt(200))
    coef_dir = directional_coefficient(np.array([1, 1]))

    gau_coef_list = []
    for i in range(1000):
        # Randomly generate the parameters for gaussian coefficient
        center = np.random.uniform(-10, 10, 2)
        amplitude = np.random.uniform(1, 2)
        sign = np.random.choice([-1, 1])
        amplitude *= sign
        sigma1 = np.random.uniform(0.5, 5)
        sigma2 = np.random.uniform(0.5, 5)
        cov = np.random.uniform(-np.sqrt(sigma1 * sigma2), np.sqrt(sigma1 * sigma2))
        sigma = np.array([[sigma1, cov], [cov, sigma2]])

        coef_gau = gaussian_coefficient(center, sigma, amplitude=amplitude)
        gau_coef_list.append(coef_gau)

    coef_gau = coefficient_wrapper(np.sum, *gau_coef_list)
    coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_dir, coef_gau)

    return coef_sum


def coef_auto_gau_strong():
    coef_radial = radial_coefficient(np.array([0, 0]), 1 / np.sqrt(200))
    coef_dir = directional_coefficient(np.array([1, 1]))

    gau_coef_list = []
    for _ in range(1000):
        # Randomly generate the parameters for gaussian coefficient
        center = np.random.uniform(-10, 10, 2)
        amplitude = np.random.uniform(1, 2)
        sign = np.random.choice([-1, 1])
        amplitude *= sign
        sigma1 = np.random.uniform(0.2, 1)
        sigma2 = np.random.uniform(0.2, 1)
        cov = np.random.uniform(-np.sqrt(sigma1 * sigma2), np.sqrt(sigma1 * sigma2))
        sigma = np.array([[sigma1, cov], [cov, sigma2]])

        coef_gau = gaussian_coefficient(center, sigma, amplitude=amplitude)
        gau_coef_list.append(coef_gau)

    coef_gau = coefficient_wrapper(np.sum, *gau_coef_list)
    coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_dir, coef_gau)

    coef_sum = coefficient_wrapper(partial(np.multiply, 2), coef_sum)

    return coef_sum


def f_square_2(X, C, points):
    return (
        polynomial_function(C[0], 2)(X[:, 0], points)
        + polynomial_function(C[1], 2)(X[:, 1], points)
        + 0
    )


def generate_sample(count, f, coef_func, random_seed=1, plot=False):
    np.random.seed(random_seed)
    points = sample_points(count, bounds=(-10, 10))
    x1 = sample_x(count, bounds=(-10, 10))
    x2 = sample_x(count, bounds=(-10, 10))

    if isinstance(coef_func, list):
        coefficients = [func() for func in coef_func]
    else:
        coefficients = [coef_func()]

    X = np.stack((x1, x2), axis=-1)
    y = f(X, coefficients, points)

    if plot:
        show_sample(X, y, points, coefficients)

    return X, y, points


def draw_graph():
    X, y, points = generate_sample(
        count=5000, f=f_square_2, coef_func=[coef_auto_gau_strong, coef_auto_gau_weak], random_seed=1,
        plot=True
    )
    X_plus = np.concatenate([X, points], axis=1)
    distance_measure = "euclidean"
    kernel_type = "bisquare"
    neighbour_count = 0.02

    model = WeightModel(
        RandomForestRegressor(n_estimators=50),
        distance_measure,
        kernel_type,
        neighbour_count=neighbour_count,
        cache_data=True,
        cache_estimator=True,
    )
    model.fit(X, y, [points])
    print("GRF:", model.llocv_score_)

    importance_global = model.importance_score_global()
    print(importance_global)

    importance_local = model.importance_score_local()
    print(importance_local)

    # Normalize the local importance to [0, 1]
    # importance_local = (importance_local - importance_local.min(axis=0)) / (
    #     importance_local.max(axis=0) - importance_local.min(axis=0)
    # )
    importance_local = (importance_local - importance_local.min(axis=1)) / (
        importance_local.max(axis=1) - importance_local.min(axis=1)
    )

    # Plot the local importance
    for i in range(importance_local.shape[1]):
        fig = plt.figure()
        scatter = plt.scatter(
            points[:, 0], points[:, 1], c=importance_local[:, i], cmap="viridis"
        )
        fig.colorbar(scatter)
        fig.savefig(f"Plot/Local_importance_{i}.png")
        fig.show()


def fit_stacking():
    X, y, points = generate_sample(
        count=5000, f=f_square_2, coef_func=[coef_auto_gau_strong, coef_auto_gau_weak], random_seed=1,
        plot=True
    )
    X_plus = np.concatenate([X, points], axis=1)
    distance_measure = "euclidean"
    kernel_type = "bisquare"

    model = StackingWeightModel(
        # ExtraTreesRegressor(n_estimators=10, max_depth=X.shape[1]),
        DecisionTreeRegressor(splitter="random", max_depth=X.shape[1]),
        distance_measure,
        kernel_type,
        neighbour_count=0.02,
        neighbour_leave_out_rate=0.2,
    )
    t1 = time.time()
    model.fit(X_plus, y, [points])
    t2 = time.time()
    print("Stacking:", model.llocv_score_, model.llocv_stacking_)
    print(t2 - t1)


if __name__ == "__main__":
    draw_graph()
    # fit_stacking()



from numba import njit, prange
from numba.typed import List
import numpy as np


@njit()
def test_stack():
    array = np.ones((2, 3))
    list_of_array = [array] * 10
    np.stack(list_of_array)


@njit()
def test_list_stack(i, array_to_be_stacked):
    shape = (i,) + array_to_be_stacked.shape
    list_of_array = [array_to_be_stacked] * i
    stacked_array = np.empty(shape)
    for j in prange(i):
        stacked_array[j] = list_of_array[j]
    return stacked_array


@njit()
def stack(list_of_array):
    shape = (len(list_of_array),) + list_of_array[0].shape
    stacked_array = np.empty(shape)
    for j in prange(len(list_of_array)):
        stacked_array[j] = list_of_array[j]
    return stacked_array

if __name__ == "__main__":
    test_stack()

    test_list_stack(10, np.ones((2, 3)))

    # Note that you have to use typed list provided by numba here.
    typed_list = List()
    [typed_list.append(np.ones((2, 3))) for _ in range(10)]
    stacked = stack(typed_list)
    print(stacked.shape)
    print(stacked)



def test_pd_sample():
    from georegression.test.visualize import get_toy_model
    model = get_toy_model()

    from georegression.visualize.pd import sample_partial
    sample_partial(
        model.feature_partial_[0], quantile=[0.1, 0.5]
    )

    from georegression.visualize.pd import partial_cluster
    _, cluster_label, _ = partial_cluster(model.feature_partial_[0])
    sample_partial(
        model.feature_partial_[0], sample_size=0.1, cluster_label=cluster_label
    )


if __name__ == '__main__':
    test_pd_sample()



from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.cluster import hierarchy
from scipy.spatial.distance import squareform
from scipy.stats import spearmanr
from sklearn.linear_model import LinearRegression
from statsmodels.stats.outliers_influence import variance_inflation_factor

from georegression.test.data import load_HP
from georegression.weight_model import WeightModel

(X, y_true, xy_vector, time) = load_HP()

# TODO: Add labels
labels = np.arange(X.shape[1])


def test_spearman(threshold=1, draw=False):
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(25, 20))
    corr = spearmanr(X).correlation

    # Ensure the correlation matrix is symmetric
    corr = (corr + corr.T) / 2
    np.fill_diagonal(corr, 1)

    # We convert the correlation matrix to a distance matrix before performing
    # hierarchical clustering using Ward's linkage.
    distance_matrix = 1 - np.abs(corr)
    dist_linkage = hierarchy.ward(squareform(distance_matrix))
    # dist_linkage = hierarchy.single(squareform(distance_matrix))
    dendro = hierarchy.dendrogram(
        dist_linkage, labels=labels, ax=ax1, leaf_rotation=90
    )
    dendro_idx = np.arange(0, len(dendro["ivl"]))

    ims = ax2.imshow(corr[dendro["leaves"], :][:, dendro["leaves"]], cmap='PiYG', vmin=-1, vmax=1)
    fig.colorbar(ims, ax=ax2)
    ax2.set_xticks(dendro_idx)
    ax2.set_yticks(dendro_idx)
    ax2.set_xticklabels(dendro["ivl"], rotation="vertical")
    ax2.set_yticklabels(dendro["ivl"])
    fig.tight_layout()
    if draw:
        plt.show()
    plt.savefig(f'test_corr.png')
    plt.clf()

    cluster_ids = hierarchy.fcluster(dist_linkage, threshold, criterion="distance")
    cluster_id_to_feature_ids = defaultdict(list)
    for idx, cluster_id in enumerate(cluster_ids):
        cluster_id_to_feature_ids[cluster_id].append(idx)
    selected_features = [v[0] for v in cluster_id_to_feature_ids.values()]

    return selected_features


def test_vif(threshold=10, output=False):
    remove_feature_list = []
    while True:
        select_feature_list = np.full(X.shape[1], True)
        select_feature_list[remove_feature_list] = False
        X_selected = X[:, select_feature_list]
        vif_list = [variance_inflation_factor(X_selected, i) for i in range(X_selected.shape[1])]

        # Inf
        # inf_index_list = np.nonzero(np.isinf(vif_list))
        # if inf_index_list[0].shape[0] != 0:
        #     remove_feature_list.append(
        #         np.where(select_feature_list)[inf_index_list[0][0]]
        #     )

        if max(vif_list) < threshold:
            if output:
                print(list(zip(vif_list, labels[select_feature_list])))
            break

        # Remove Max VIF Feature
        vif_index = vif_list.index(max(vif_list))
        remove_feature_list.append(
            np.where(select_feature_list)[0][vif_index]
        )

    return select_feature_list


def test_customer_vif():
    from sklearn.linear_model import LinearRegression
    estimator = LinearRegression()

    feature_index = 12

    estimator.fit(X[:, [*range(feature_index), *range(feature_index + 1, X.shape[1])]], X[:, feature_index])
    r2 = estimator.score(X[:, [*range(feature_index), *range(feature_index + 1, X.shape[1])]], X[:, feature_index])
    VIF = 1 / (1 - r2)
    print(estimator.coef_, r2, VIF)


def test_tree_based_collinear(threshold=0.5, output=False):
    """

    May not work.

    """

    from rfpimp import oob_dependences
    from sklearn.ensemble import RandomForestRegressor
    estimator = RandomForestRegressor(oob_score=True)

    # Not linear regression, not VIF
    # vif_list = 1 / (1 - df_dep.values)

    remove_feature_list = []
    while True:
        select_feature_list = np.full(X.shape[1], True)
        select_feature_list[remove_feature_list] = False
        X_selected = X[:, select_feature_list]
        df_dep = oob_dependences(estimator, pd.DataFrame(X).drop_duplicates(subset=[37, 38])).sort_index()
        score_list = df_dep.values.flatten().tolist()

        if max(score_list) < threshold:
            if output:
                print(list(zip(score_list, labels[select_feature_list])))
            break

        # Remove Max Score Feature
        max_index = score_list.index(max(score_list))
        remove_feature_list.append(
            np.where(select_feature_list)[0][max_index]
        )

    return select_feature_list


def test_select_feature():
    for threshold in np.arange(0.1, 1, 0.1):
        select_features = test_tree_based_collinear(threshold)
        X_selected = X[:, select_features]

        estimator = WeightModel(
            LinearRegression(),
            distance_measure='euclidean',
            kernel_type='bisquare',
            neighbour_count=0.1,

            cache_data=True, cache_estimator=True
        )
        estimator.fit(X_selected, y_true, xy_vector, time)
        print(f'Score {threshold}')
        print(f'Feature Num: {np.nonzero(select_features)[0].shape[0]}')
        print(f'Feature Name: {labels[select_features]}')
        print(estimator.llocv_score_)
        print()



"""
Generate simulated data for testing purposes.
A Bayesian Implementation of the Multiscale Geographically Weighted Regression Model with INLA
https://doi.org/10.1080/24694452.2023.2187756
"""
from functools import partial
import matplotlib.pyplot as plt
import numpy as np

from georegression.simulation.simulation_utils import (
    gaussian_coefficient,
    interaction_function,
    radial_coefficient,
    directional_coefficient,
    sine_coefficient,
    coefficient_wrapper,
    polynomial_function,
    sigmoid_function,
    sample_points,
    sample_x,
)


def f_interact(X, C, points):
    return interaction_function(C[0])(X[:, 0], X[:, 1], points) + 0


def f_square_2(X, C, points):
    return (
        polynomial_function(C[0], 2)(X[:, 0], points)
        + polynomial_function(C[1], 2)(X[:, 1], points)
        + 0
    )


def coef_strong():
    coef_radial = radial_coefficient(np.array([0, 0]), 1 / np.sqrt(200))
    coef_dir = directional_coefficient(np.array([1, 1]))
    coef_sin_1 = sine_coefficient(1, np.array([-1, 1]), 1)
    coef_sin_2 = sine_coefficient(1, np.array([1, 1]), 1)
    coef_sin = coefficient_wrapper(np.sum, coef_sin_1, coef_sin_2)
    coef_gau_1 = gaussian_coefficient(np.array([-5, 5]), 3)
    coef_gau_2 = gaussian_coefficient(np.array([-5, -5]), 3, amplitude=2)
    coef_gau = coefficient_wrapper(np.sum, coef_gau_1, coef_gau_2)

    coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_dir, coef_sin, coef_gau)

    return coef_sum


def coef_manual_gau():
    coef_radial = radial_coefficient(np.array([0, 0]), 1 / np.sqrt(200))
    coef_dir = directional_coefficient(np.array([1, 1]))

    coef_gau_1 = gaussian_coefficient(np.array([-5, 5]), [[3, 4], [4, 8]], amplitude=-1)
    coef_gau_2 = gaussian_coefficient(np.array([-2, -5]), 5, amplitude=2)
    coef_gau_3 = gaussian_coefficient(np.array([8, 3]), 10, amplitude=-1.5)
    coef_gau_4 = gaussian_coefficient(
        np.array([2, 8]), [[3, 0], [0, 15]], amplitude=0.8
    )
    coef_gau_5 = gaussian_coefficient(np.array([5, -10]), 1, amplitude=1)
    coef_gau_6 = gaussian_coefficient(np.array([-10, -10]), 15, amplitude=1.5)
    coef_gau_6 = gaussian_coefficient(np.array([-11, 0]), 5, amplitude=2)
    coef_gau_6 = gaussian_coefficient(np.array([-11, 0]), 5, amplitude=2)
    coef_gau = coefficient_wrapper(
        np.sum, coef_gau_1, coef_gau_2, coef_gau_3, coef_gau_4, coef_gau_5, coef_gau_6
    )

    # coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_dir, coef_gau)
    coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_gau)

    return coef_sum


def coef_auto_gau_weak():
    coef_radial = radial_coefficient(np.array([0, 0]), 1 / np.sqrt(200))
    coef_dir = directional_coefficient(np.array([1, 1]))

    gau_coef_list = []
    for i in range(1000):
        # Randomly generate the parameters for gaussian coefficient
        center = np.random.uniform(-10, 10, 2)
        amplitude = np.random.uniform(1, 2)
        sign = np.random.choice([-1, 1])
        amplitude *= sign
        sigma1 = np.random.uniform(0.5, 5)
        sigma2 = np.random.uniform(0.5, 5)
        cov = np.random.uniform(-np.sqrt(sigma1 * sigma2), np.sqrt(sigma1 * sigma2))
        sigma = np.array([[sigma1, cov], [cov, sigma2]])

        coef_gau = gaussian_coefficient(center, sigma, amplitude=amplitude)
        gau_coef_list.append(coef_gau)

    coef_gau = coefficient_wrapper(np.sum, *gau_coef_list)
    coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_dir, coef_gau)

    return coef_sum


def coef_auto_gau_strong():
    coef_radial = radial_coefficient(np.array([0, 0]), 1 / np.sqrt(200))
    coef_dir = directional_coefficient(np.array([1, 1]))

    gau_coef_list = []
    for _ in range(1000):
        # Randomly generate the parameters for gaussian coefficient
        center = np.random.uniform(-10, 10, 2)
        amplitude = np.random.uniform(1, 2)
        sign = np.random.choice([-1, 1])
        amplitude *= sign
        sigma1 = np.random.uniform(0.2, 1)
        sigma2 = np.random.uniform(0.2, 1)
        cov = np.random.uniform(-np.sqrt(sigma1 * sigma2), np.sqrt(sigma1 * sigma2))
        sigma = np.array([[sigma1, cov], [cov, sigma2]])

        coef_gau = gaussian_coefficient(center, sigma, amplitude=amplitude)
        gau_coef_list.append(coef_gau)

    coef_gau = coefficient_wrapper(np.sum, *gau_coef_list)
    coef_sum = coefficient_wrapper(np.sum, coef_radial, coef_dir, coef_gau)

    return coef_sum


f = f_interact
coef_func = coef_manual_gau
x2_coef = coefficient_wrapper(partial(np.multiply, 3), coef_func())


def generate_sample(random_seed=None, count=100, f=f, coef_func=coef_func):
    np.random.seed(random_seed)

    points = sample_points(count, bounds=[[-10, 10], [-10, 10]])

    # x1 = sample_x(count)
    x1 = sample_x(count, mean=coef_func(), bounds=(-1, 1), points=points)

    # x2 = sample_x(count)
    # x2 = sample_x(count, bounds=(0, 1))
    x2_coef = coefficient_wrapper(partial(np.multiply, 3), coef_func())
    x2 = sample_x(count, mean=x2_coef, bounds=(-2, 2), points=points)

    if isinstance(coef_func, list):
        coefficients = [func() for func in coef_func]
    else:
        coefficients = [coef_func()]

    X = np.stack((x1, x2), axis=-1)
    y = f(X, coefficients, points)

    return X, y, points, f, coefficients


def show_sample(X, y, points, coefficients, folder="Plot"):
    """
    Show X, y, points, and coefficients in multiple subplots.
    Assume dimension of points is 2, which is a plane.
    """

    dim_points = points.shape[1]
    if dim_points != 2:
        raise ValueError("Dimension of points must be 2.")

    # Calculate the number of subplots needed.
    dim_x = X.shape[1]
    dim_coef = len(coefficients)

    # Plot X. Add colorbar for each dimension.
    plt.figure()
    for i in range(dim_x):
        plt.figure()
        plt.scatter(points[:, 0], points[:, 1], c=X[:, i], cmap="Spectral")
        plt.colorbar()
        plt.xlabel("x")
        plt.ylabel("y")
        plt.title(f"The {i+1}-th feature of X")
        plt.savefig(f"{folder}/Simulation_X_{i+1}.png")

    # Plot y using scatter and boxplot
    plt.figure()
    plt.scatter(points[:, 0], points[:, 1], c=y, cmap="Spectral")
    plt.colorbar()
    plt.title("The value of y across the plane")
    plt.savefig(f"{folder}/Simulation_y.png")

    plt.figure()
    plt.boxplot(y)
    plt.title("The distribution of y")
    plt.savefig(f"{folder}/Simulation_y_boxplot.png")

    # Plot coefficients
    for i in range(dim_coef):
        plt.figure()
        plt.scatter(
            points[:, 0], points[:, 1], c=coefficients[i](points), cmap="Spectral"
        )
        plt.colorbar()
        plt.title(f"The {i+1}-th coefficient across the plane")
        plt.savefig(f"{folder}/Simulation_Coefficients_{i}.png")


def main():
    X, y, points, f, coefficients = generate_sample(count=5000, random_seed=1)

    show_sample(X, y, points, coefficients)
    plt.show(block=True)

    # Save all figure to the local directory
    # for i in plt.get_fignums():
    # plt.figure(i)
    # plt.savefig(f"figure{i}.png")


if __name__ == "__main__":
    main()



from time import time
from typing import Union

import dask.array as da
import numpy as np
from dask.graph_manipulation import wait_on
from scipy import sparse
from slab_utils.quick_logger import logger

from georegression.distance_utils import _distance_matrices
from georegression.kernel import kernel_function, adaptive_kernel


def weight_matrix_from_points(
        source_coordinate_vector_list: list[np.ndarray],
        target_coordinate_vector_list: list[np.ndarray] = None,
        distance_measure: Union[str, list[str]] = None,
        kernel_type: Union[str, list[str]] = None,
        distance_ratio: Union[float, list[float]] = None,
        bandwidth: Union[float, list[float]] = None,
        neighbour_count: Union[float, list[float]] = None,
        distance_args: Union[dict, list[dict]] = None
) -> np.ndarray:
    """
    Iterate over each source-target pair to get weight matrix.
    Each row represent each source. Each column represent each target.
    The shape of the matrix is (number of source, number of target).

    Args:
        source_coordinate_vector_list:
        target_coordinate_vector_list:
        distance_measure:
        kernel_type:
        distance_ratio:
        bandwidth:
        neighbour_count:
        distance_args:

    Returns:

    """

    log_str = f"\nWeight Matrix from Points Info:\n"
    log_str += f"Coordinate Dimension: {len(source_coordinate_vector_list)}\n"
    for i, coordinate_vector in enumerate(source_coordinate_vector_list):
        log_str += f"coordinate_vector[{i}].shape: {coordinate_vector.shape}\n"
    logger.debug(log_str)

    t_distance_start = time()
    distance_matrices = _distance_matrices(
        source_coordinate_vector_list,
        target_coordinate_vector_list,
        distance_measure,
        distance_args,
    )
    t_distance_end = time()

    t_kernel_start = time()
    compound_weight_matrix = weight_matrix_from_distance(
        distance_matrices, kernel_type, distance_ratio, bandwidth, neighbour_count
    )
    t_kernel_end = time()

    logger.debug(f"Distance Time: {t_distance_end - t_distance_start}. Kernel Time: {t_kernel_end - t_kernel_start}")

    return compound_weight_matrix


def weight_matrix_from_distance(
        distance_matrices,
        kernel_type: Union[str, list[str]],
        distance_ratio: Union[float, list[float], None] = None,
        bandwidth: Union[float, list[float], None] = None,
        neighbour_count: Union[float, list[float], None] = None,
        distance_matrices_sorted: Union[np.ndarray, da.array] = None
) -> Union[np.ndarray, da.array]:
    """
    Calculate weights for each coordinate vector (e.g. location coordinate vector or time coordinate vector)
    and integrate the weights of each coordinate vector to one weight
    using some arithmetic operations (e.g. add or multiply).

    Or in the reversed order, Integrate the distances of each coordinate vector and calculate the weight.
    In this case, `distance_ratio` should be provided.

    All the parameters can provide in list form if weights are integrated instead of distance.
    Length of the lists should match the dimension(or length) of the vector list.

    Args:
        one_coordinate_vector_list:
        many_coordinate_vector_list:
        distance_measure:
        kernel_type:
        distance_ratio:
        bandwidth:
        neighbour_count:
        p:

    Returns:

    """

    log_str = f"\nWeight Matrix from Distance Info:\n"
    log_str += f"Distance Dimension: {len(distance_matrices)}\n"
    for i, distance_matrix in enumerate(distance_matrices):
        log_str += f"distance_matrix[{i}].shape: {distance_matrix.shape}\n"
    log_str += (
        f"kernel_type: {kernel_type}\n"
        f"distance_ratio: {distance_ratio}\n"
        f"bandwidth: {bandwidth}\n"
        f"neighbour_count: {neighbour_count}\n"
    )
    logger.debug(log_str)

    # Dimension of the vector list. (Len of the vector list)
    dimension = len(distance_matrices)

    # Check whether the size of distance matrices are the same.
    if len(set([distance_matrix.shape for distance_matrix in distance_matrices])) != 1:
        raise Exception("Size of distance matrices are not the same")

    # Check whether to use fixed kernel or adaptive kernel
    if bandwidth is None and neighbour_count is None:
        raise Exception(
            "At least one of bandwidth or neighbour count should be provided"
        )

    # Merge distance matrices.
    if distance_ratio is not None:
        if not isinstance(distance_ratio, list) and dimension != 2:
            raise Exception(
                "Distance ratio list must be provided for dimension larger than 2"
            )

        # TODO: Normalization step should be considered.

        # TODO: More operation, not only addition, should be considered.
        #  Like different distance measurements (replace distance_diff in distance_utils.py).
        #  Or some arithmetic operations like multiplication or division?

        if not isinstance(distance_ratio, list) and dimension == 2:
            distance_ratio = [1, distance_ratio]
        else:
            if len(distance_ratio) != dimension:
                raise Exception(
                    "Length of distance ratio list must match the dimension of the vector list"
                )

        distance_matrices_temp = []
        for dim in range(dimension):
            distance_matrices_temp[dim] = distance_matrices[dim] * distance_ratio[dim]

        distance_matrix = np.sum(distance_matrices_temp, axis=0)
        distance_matrices = [distance_matrix]

        dimension = 1

    # Then, calculate weight matrices and merge.

    # Also should check the dimension of the parameters if it is already a list?
    if not isinstance(kernel_type, list):
        kernel_type = [kernel_type] * dimension

    if not isinstance(bandwidth, list):
        bandwidth = [bandwidth] * dimension

    if not isinstance(neighbour_count, list):
        neighbour_count = [neighbour_count] * dimension

    if not isinstance(distance_matrices_sorted, list):
        distance_matrices_sorted = [distance_matrices_sorted] * dimension

    weights = []
    for dim in range(dimension):
        if isinstance(distance_matrices[0], da.Array):
            weights.append(
                # Need to wait on?
                weight_by_distance(
                    distance_matrices[dim], kernel_type[dim], bandwidth[dim], neighbour_count[dim],
                    distance_matrices_sorted[dim]
                )
            )
        else:
            weights.append(
                weight_by_distance(distance_matrices[dim], kernel_type[dim], bandwidth[dim], neighbour_count[dim], None)
            )

    weights = np.stack(weights)

    # TODO: Not only multiplication? e.g. Addition, minimum, maximum, average
    weight_matrix = np.prod(weights, axis=0)

    # Normalization
    # TODO: More normalization option. The key point is the proportion in a row?

    # default use row normalization
    row_sum = np.sum(weight_matrix, axis=1)
    # for some row with all 0 weight.
    row_sum[row_sum == 0] = 1
    # Notice the axis of division
    weight_matrix_norm = weight_matrix / np.expand_dims(row_sum, 1)

    if isinstance(weight_matrix_norm, da.Array):
        weight_matrix_norm = weight_matrix_norm.map_blocks(sparse.coo_matrix).compute()
        weight_matrix_norm = sparse.csr_array(weight_matrix_norm)

    # Stat the non-zero weight ratio
    if isinstance(weight_matrix_norm, np.ndarray):
        nonzero_count = np.count_nonzero(weight_matrix_norm)
    else:
        nonzero_count = weight_matrix_norm.nnz

    logger.debug(
        f"Non-zero weight ratio: {nonzero_count / weight_matrix_norm.size}\n"
        f"Average neighbour count: {nonzero_count / weight_matrix_norm.shape[0]}"
    )

    return weight_matrix_norm


def weight_by_distance(distance, kernel_type, bandwidth, neighbour_count, distance_sorted):
    """
    Using fixed kernel(bandwidth provided) or adaptive kernel(neighbour count provided)
    to calculate the weight based on the distance vector.

    Args:
        distance:
        kernel_type:
        bandwidth:
        neighbour_count:

    Returns:

    """

    if bandwidth is not None and neighbour_count is None:
        weight = kernel_function(distance, bandwidth, kernel_type)
    elif bandwidth is None and neighbour_count is not None:
        weight = adaptive_kernel(distance, neighbour_count, kernel_type, distance_sorted)
    else:
        raise Exception(
            "Choose bandwidth for fixed kernel or neighbour count for adaptive kernel"
        )
    return weight



import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestRegressor

from georegression.local_ale import weighted_ale

X = np.random.normal(size=(100, 5))
y = np.random.normal(size=100)

df = pd.DataFrame(X, columns=["x1", "x2", "x3", "x4", "x5"])

estimator = RandomForestRegressor(n_estimators=100)
estimator.fit(X, y)


def test_weighted_ale():
    weighted_ale(X, 0, estimator.predict, weights=np.random.random(size=100))

if __name__ == "__main__":
    test_weighted_ale()
